#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0xcc1d10

# Compiled with Coconut version 1.3.1 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_tail_call, _coconut_tco, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

import builtins  # line 1
import json  # line 1
import logging  # line 1
import os  # line 1
import shutil  # line 1
sys = _coconut_sys  # line 1
import time  # line 1
import traceback  # line 1
import unittest  # line 1
import uuid  # line 1
StringIO = (__import__("StringIO" if sys.version_info.major < 3 else "io")).StringIO  # enables import via ternary expression  # line 2
try:  # Py3  # line 3
    from unittest import mock  # Py3  # line 3
except:  # installed via pip  # line 4
    import mock  # installed via pip  # line 4
try:  # only required for mypy  # line 5
    from typing import Any  # only required for mypy  # line 5
    from typing import List  # only required for mypy  # line 5
    from typing import Union  # only required for mypy  # line 5
except:  # line 6
    pass  # line 6

testFolder = os.path.abspath(os.path.join(os.getcwd(), "test"))  # line 8
try:  # line 9
    import configr  # optional dependency  # line 10
    os.environ["TEST"] = testFolder  # needed to mock configr library calls in sos  # line 11
except:  # declare as undefined  # line 12
    configr = None  # declare as undefined  # line 12
import sos  # line 13
sos.defaults["defaultbranch"] = "trunk"  # because sos.main() is never called  # line 14

def sync() -> 'None':  # line 16
    if (sys.version_info.major, sys.version_info.minor) >= (3, 3):  # line 17
        os.sync()  # line 17


def determineFilesystemTimeResolution() -> 'float':  # line 20
    name = str(uuid.uuid4())  # line 21
    with open(name, "w") as fd:  # create temporary file  # line 22
        fd.write("x")  # create temporary file  # line 22
    mt = os.stat(name)[8]  # get current timestamp  # line 23
    while os.stat(name)[8] == mt:  # wait until timestamp modified  # line 24
        with open(name, "w") as fd:  # line 25
            fd.write("x")  # line 25
    mt, start, count = os.stat(name)[8], time.time(), 0  # line 26
    while os.stat(name)[8] == mt:  # now cound and measure time until modified again  # line 27
        count += 1  # line 28
        with open(name, "w") as fd:  # line 29
            fd.write("x")  # line 29
    os.unlink(name)  # line 30
    fsprecision = round(time.time() - start, 2)  # line 31
    print("File system timestamp precision is %.2fs; wrote to the file %d times during that time" % (fsprecision, count))  # line 32
    return fsprecision  # line 33


FS_PRECISION = determineFilesystemTimeResolution() * 1.05  # line 36


@_coconut_tco  # line 39
def debugTestRunner(post_mortem=None):  # line 39
    ''' Unittest runner doing post mortem debugging on failing tests. '''  # line 40
    import pdb  # line 41
    if post_mortem is None:  # line 42
        post_mortem = pdb.post_mortem  # line 42
    class DebugTestResult(unittest.TextTestResult):  # line 43
        def addError(self, test, err):  # called before tearDown()  # line 44
            traceback.print_exception(*err)  # line 45
            post_mortem(err[2])  # line 46
            super(DebugTestResult, self).addError(test, err)  # line 47
        def addFailure(self, test, err):  # line 48
            traceback.print_exception(*err)  # line 49
            post_mortem(err[2])  # line 50
            super(DebugTestResult, self).addFailure(test, err)  # line 51
    return _coconut_tail_call(unittest.TextTestRunner, resultclass=DebugTestResult)  # line 52

def branchFolder(branch: 'int', revision: 'int') -> 'str':  # line 54
    return "." + os.sep + sos.metaFolder + os.sep + "b%d" % branch + os.sep + "r%d" % revision  # line 54

@_coconut_tco  # line 56
def wrapChannels(func: '_coconut.typing.Callable[..., Any]'):  # line 56
    ''' Wrap function call to capture and return strings emitted on stdout and stderr. '''  # line 57
    oldv, oldo, olde = sys.argv, sys.stdout, sys.stderr  # line 58
    buf = StringIO()  # line 59
    sys.stdout = sys.stderr = buf  # line 60
    handler = logging.StreamHandler(buf)  # line 61
    logging.getLogger().addHandler(handler)  # line 62
    try:  # capture output into buf  # line 63
        func()  # capture output into buf  # line 63
    except Exception as E:  # line 64
        buf.write(str(E) + "\n")  # line 64
        traceback.print_exc(file=buf)  # line 64
    logging.getLogger().removeHandler(handler)  # line 65
    sys.argv, sys.stdout, sys.stderr = oldv, oldo, olde  # line 66
    return _coconut_tail_call(buf.getvalue)  # line 67

def mockInput(datas: '_coconut.typing.Sequence[str]', func) -> 'Any':  # line 69
    with mock.patch("builtins.input" if sys.version_info.major >= 3 else "utility._coconut_raw_input", side_effect=datas):  # line 70
        return func()  # line 70

def setRepoFlag(name: 'str', value: 'bool') -> 'None':  # line 72
    with open(sos.metaFolder + os.sep + sos.metaFile, "r") as fd:  # line 73
        flags, branches = json.loads(fd.read())  # line 73
    flags[name] = value  # line 74
    with open(sos.metaFolder + os.sep + sos.metaFile, "w") as fd:  # line 75
        fd.write(json.dumps((flags, branches)))  # line 75

def checkRepoFlag(name: 'str', flag: 'bool') -> 'bool':  # line 77
    with open(sos.metaFolder + os.sep + sos.metaFile, "r") as fd:  # line 77
        flags, branches = json.loads(fd.read())  # line 78
    return name in flags and flags[name] == flag  # line 79


class Tests(unittest.TestCase):  # line 82
    ''' Entire test suite. '''  # line 83

    def setUp(_):  # line 85
        for entry in os.listdir(testFolder):  # cannot remove testFolder on Windows when using TortoiseSVN as VCS  # line 86
            resource = os.path.join(testFolder, entry)  # line 87
            shutil.rmtree(resource) if os.path.isdir(resource) else os.unlink(resource)  # line 88
        os.chdir(testFolder)  # line 89

    def assertAllIn(_, what: '_coconut.typing.Sequence[str]', where: 'Union[str, List[str]]') -> 'None':  # line 91
        [_.assertIn(w, where) for w in what]  # line 91

    def assertInAll(_, what: 'str', where: '_coconut.typing.Sequence[str]') -> 'None':  # line 93
        [_.assertIn(what, w) for w in where]  # line 93

    def assertInAny(_, what: 'str', where: '_coconut.typing.Sequence[str]') -> 'None':  # line 95
        _.assertTrue(any((what in w for w in where)))  # line 95

    def assertNotInAny(_, what: 'str', where: '_coconut.typing.Sequence[str]') -> 'None':  # line 97
        _.assertFalse(any((what in w for w in where)))  # line 97

    def createFile(_, number: 'Union[int, str]', contents: 'str'="x" * 10, prefix: '_coconut.typing.Optional[str]'=None) -> 'None':  # line 99
        with open(("." if prefix is None else prefix) + os.sep + (("file%d" % number) if isinstance(number, int) else number), "wb") as fd:  # line 100
            fd.write(contents if isinstance(contents, bytes) else contents.encode("cp1252"))  # line 100

    def existsFile(_, number: 'Union[int, str]', expectedContents: 'bytes'=None) -> 'bool':  # line 102
        if not os.path.exists(("." + os.sep + "file%d" % number) if isinstance(number, int) else number):  # line 103
            return False  # line 103
        if expectedContents is None:  # line 104
            return True  # line 104
        with open(("." + os.sep + "file%d" % number) if isinstance(number, int) else number, "rb") as fd:  # line 105
            return fd.read() == expectedContents  # line 105

    def testAccessor(_):  # line 107
        a = sos.Accessor({"a": 1})  # line 108
        _.assertEqual((1, 1), (a["a"], a.a))  # line 109

    def testFirstofmap(_):  # line 111
        _.assertEqual(2, sos.firstOfMap({"a": 1, "b": 2}, ["x", "b"]))  # line 112
        _.assertIsNone(sos.firstOfMap({"a": 1, "b": 2}, []))  # line 113

    def testAjoin(_):  # line 115
        _.assertEqual("a1a2", sos.ajoin("a", ["1", "2"]))  # line 116
        _.assertEqual("* a\n* b", sos.ajoin("* ", ["a", "b"], "\n"))  # line 117

    def testFindChanges(_):  # line 119
        m = sos.Metadata(os.getcwd())  # line 120
        m.loadBranches()  # line 121
        _.createFile(1, "1")  # line 122
        m.createBranch(0)  # line 123
        _.assertEqual(1, len(m.paths))  # line 124
        time.sleep(FS_PRECISION)  # time required by filesystem time resolution issues  # line 125
        _.createFile(1, "2")  # line 126
        _.createFile(2, "2")  # line 127
        changes = m.findChanges()  # detect time skew  # line 128
        _.assertEqual(1, len(changes.additions))  # line 129
        _.assertEqual(0, len(changes.deletions))  # line 130
        _.assertEqual(1, len(changes.modifications))  # line 131
        m.integrateChangeset(changes)  # line 132
        _.createFile(2, "12")  # modify file  # line 133
        changes = m.findChanges(0, 1)  # by size, creating new commit  # line 134
        _.assertEqual(0, len(changes.additions))  # line 135
        _.assertEqual(0, len(changes.deletions))  # line 136
        _.assertEqual(1, len(changes.modifications))  # line 137
        _.assertTrue(os.path.exists(branchFolder(0, 1)))  # line 138
        _.assertTrue(os.path.exists(branchFolder(0, 1) + os.sep + "03b69bc801ae11f1ff2a71a50f165996d0ad681b4f822df13329a27e53f0fcd2"))  # line 139

    def testDiffFunc(_):  # line 141
        a = {"./a": sos.PathInfo("", 0, 0, "")}  # line 142
        b = {"./a": sos.PathInfo("", 0, 0, "")}  # line 143
        changes = sos.diffPathSets(a, b)  # line 144
        _.assertEqual(0, len(changes.additions))  # line 145
        _.assertEqual(0, len(changes.deletions))  # line 146
        _.assertEqual(0, len(changes.modifications))  # line 147
        b = {"./a": sos.PathInfo("", 1, 0, "")}  # line 148
        changes = sos.diffPathSets(a, b)  # line 149
        _.assertEqual(0, len(changes.additions))  # line 150
        _.assertEqual(0, len(changes.deletions))  # line 151
        _.assertEqual(1, len(changes.modifications))  # line 152
        b = {}  # diff contains no entries -> no change  # line 153
        changes = sos.diffPathSets(a, b)  # line 154
        _.assertEqual(0, len(changes.additions))  # line 155
        _.assertEqual(0, len(changes.deletions))  # line 156
        _.assertEqual(0, len(changes.modifications))  # line 157
        b = {"./a": sos.PathInfo("", None, 1, "")}  # in diff marked as deleted  # line 158
        changes = sos.diffPathSets(a, b)  # line 159
        _.assertEqual(0, len(changes.additions))  # line 160
        _.assertEqual(1, len(changes.deletions))  # line 161
        _.assertEqual(0, len(changes.modifications))  # line 162
        b = {"./b": sos.PathInfo("", 1, 1, "")}  # line 163
        changes = sos.diffPathSets(a, b)  # line 164
        _.assertEqual(1, len(changes.additions))  # line 165
        _.assertEqual(0, len(changes.deletions))  # line 166
        _.assertEqual(0, len(changes.modifications))  # line 167
        a = {"./a": sos.PathInfo("", None, 0, "")}  # mark as deleted  # line 168
        b = {"./a": sos.PathInfo("", 1, 0, "")}  # re-added  # line 169
        changes = sos.diffPathSets(a, b)  # line 170
        _.assertEqual(1, len(changes.additions))  # line 171
        _.assertEqual(0, len(changes.deletions))  # line 172
        _.assertEqual(0, len(changes.modifications))  # line 173
        changes = sos.diffPathSets(b, a)  # line 174
        _.assertEqual(0, len(changes.additions))  # line 175
        _.assertEqual(1, len(changes.deletions))  # line 176
        _.assertEqual(0, len(changes.modifications))  # line 177

    def testPatternPaths(_):  # line 179
        a = False  # type: bool  # line 180
        def b():  # line 181
            nonlocal a  # line 181
            a = True  # line 181
        co = sos.CallOnce(b)  # type: sos.CallOnce  # line 182
        co()  # line 183
        _.assertTrue(a)  # was modified  # line 184
        a = False  # line 185
        co()  # line 186
        _.assertFalse(a)  # was not modified again  # line 187
        sos.offline(options=["--track"])  # line 188
        os.mkdir("sub")  # line 189
        _.createFile("sub" + os.sep + "file1", "sdfsdf")  # line 190
        sos.add("./sub", "sub/file?")  # line 191
        sos.commit("test")  # should pick up sub/file1 pattern  # line 192
        _.assertEqual(2, len(os.listdir(os.path.join(sos.metaFolder, "b0", "r1"))))  # sub/file1 was added  # line 193
        _.createFile(1)  # line 194
        try:  # should not commit anything, as the file in base folder doesn't match the tracked pattern  # line 195
            sos.commit("nothing")  # should not commit anything, as the file in base folder doesn't match the tracked pattern  # line 195
            _.fail()  # should not commit anything, as the file in base folder doesn't match the tracked pattern  # line 195
        except:  # line 196
            pass  # line 196

    def testComputeSequentialPathSet(_):  # line 198
        os.makedirs(branchFolder(0, 0))  # line 199
        os.makedirs(branchFolder(0, 1))  # line 200
        os.makedirs(branchFolder(0, 2))  # line 201
        os.makedirs(branchFolder(0, 3))  # line 202
        os.makedirs(branchFolder(0, 4))  # line 203
        m = sos.Metadata(os.getcwd())  # line 204
        m.branch = 0  # line 205
        m.commit = 2  # line 206
        m.saveBranches()  # line 207
        m.paths = {"./a": sos.PathInfo("", 0, 0, "")}  # line 208
        m.saveCommit(0, 0)  # initial  # line 209
        m.paths["./a"] = sos.PathInfo("", 1, 0, "")  # line 210
        m.saveCommit(0, 1)  # mod  # line 211
        m.paths["./b"] = sos.PathInfo("", 0, 0, "")  # line 212
        m.saveCommit(0, 2)  # add  # line 213
        m.paths["./a"] = sos.PathInfo("", None, 0, "")  # line 214
        m.saveCommit(0, 3)  # del  # line 215
        m.paths["./a"] = sos.PathInfo("", None, 0, "")  # line 216
        m.saveCommit(0, 4)  # readd  # line 217
        m.commits = {i: sos.CommitInfo(i, 0, None) for i in range(5)}  # line 218
        m.saveBranch(0)  # line 219
        m.computeSequentialPathSet(0, 4)  # line 220
        _.assertEqual(2, len(m.paths))  # line 221

    def testParseRevisionString(_):  # line 223
        m = sos.Metadata(os.getcwd())  # line 224
        m.branch = 1  # line 225
        m.commits = {0: 0, 1: 1, 2: 2}  # line 226
        _.assertEqual((1, 3), m.parseRevisionString("3"))  # line 227
        _.assertEqual((2, 3), m.parseRevisionString("2/3"))  # line 228
        _.assertEqual((1, -1), m.parseRevisionString(None))  # line 229
        _.assertEqual((2, -1), m.parseRevisionString("2/"))  # line 230
        _.assertEqual((1, -2), m.parseRevisionString("/-2"))  # line 231
        _.assertEqual((1, -1), m.parseRevisionString("/"))  # line 232

    def testOfflineEmpty(_):  # line 234
        os.mkdir("." + os.sep + sos.metaFolder)  # line 235
        try:  # line 236
            sos.offline("trunk")  # line 236
            _.fail()  # line 236
        except SystemExit:  # line 237
            pass  # line 237
        os.rmdir("." + os.sep + sos.metaFolder)  # line 238
        sos.offline("test")  # line 239
        _.assertIn(sos.metaFolder, os.listdir("."))  # line 240
        _.assertAllIn(["b0", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder))  # line 241
        _.assertAllIn(["r0", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0"))  # line 242
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder)))  # only branch folder and meta data file  # line 243
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0")))  # only commit folder and meta data file  # line 244
        _.assertEqual(1, len(os.listdir(branchFolder(0, 0))))  # only meta data file  # line 245

    def testOfflineWithFiles(_):  # line 247
        _.createFile(1, "x" * 100)  # line 248
        _.createFile(2)  # line 249
        sos.offline("test")  # line 250
        _.assertAllIn(["file1", "file2", sos.metaFolder], os.listdir("."))  # line 251
        _.assertAllIn(["b0", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder))  # line 252
        _.assertAllIn(["r0", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0"))  # line 253
        _.assertAllIn([sos.metaFile, "03b69bc801ae11f1ff2a71a50f165996d0ad681b4f822df13329a27e53f0fcd2", "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa"], os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0" + os.sep + "r0"))  # line 254
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder)))  # only branch folder and meta data file  # line 255
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0")))  # only commit folder and meta data file  # line 256
        _.assertEqual(3, len(os.listdir(branchFolder(0, 0))))  # only meta data file plus branch base file copies  # line 257

    def testBranch(_):  # line 259
        _.createFile(1, "x" * 100)  # line 260
        _.createFile(2)  # line 261
        sos.offline("test")  # b0/r0  # line 262
        sos.branch("other")  # b1/r0  # line 263
        _.assertAllIn(["b0", "b1", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder))  # line 264
        _.assertEqual(list(sorted(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0"))), list(sorted(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b1"))))  # line 265
        _.assertEqual(list(sorted(os.listdir(branchFolder(0, 0)))), list(sorted(os.listdir(branchFolder(1, 0)))))  # line 267
        _.createFile(1, "z")  # modify file  # line 269
        sos.branch()  # b2/r0  branch to unnamed branch with modified file tree contents  # line 270
        _.assertNotEqual(os.stat("." + os.sep + sos.metaFolder + os.sep + "b1" + os.sep + "r0" + os.sep + "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa").st_size, os.stat("." + os.sep + sos.metaFolder + os.sep + "b2" + os.sep + "r0" + os.sep + "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa").st_size)  # line 271
        _.createFile(3, "z")  # line 273
        sos.branch("from_last_revision", ["--last", "--stay"])  # b3/r0 create copy of other file1,file2 and don't switch  # line 274
        _.assertEqual(list(sorted(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b3" + os.sep + "r0"))), list(sorted(os.listdir("." + os.sep + sos.metaFolder + os.sep + "b2" + os.sep + "r0"))))  # line 275
# Check sos.status output which branch is marked


    def testComittingAndChanges(_):  # line 280
        _.createFile(1, "x" * 100)  # line 281
        _.createFile(2)  # line 282
        sos.offline("test")  # line 283
        changes = sos.changes()  # line 284
        _.assertEqual(0, len(changes.additions))  # line 285
        _.assertEqual(0, len(changes.deletions))  # line 286
        _.assertEqual(0, len(changes.modifications))  # line 287
        _.createFile(1, "z")  # size change  # line 288
        changes = sos.changes()  # line 289
        _.assertEqual(0, len(changes.additions))  # line 290
        _.assertEqual(0, len(changes.deletions))  # line 291
        _.assertEqual(1, len(changes.modifications))  # line 292
        sos.commit("message")  # line 293
        _.assertAllIn(["r0", "r1", sos.metaFile], os.listdir("." + os.sep + sos.metaFolder + os.sep + "b0"))  # line 294
        _.assertAllIn([sos.metaFile, "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa"], os.listdir(branchFolder(0, 1)))  # line 295
        _.assertEqual(2, len(os.listdir(branchFolder(0, 1))))  # no further files, only the modified one  # line 296
        _.assertEqual(1, len(sos.changes("/0").modifications))  # vs. explicit revision on current branch  # line 297
        _.assertEqual(1, len(sos.changes("0/0").modifications))  # vs. explicit branch/revision  # line 298
        _.createFile(1, "")  # modify to empty file, mentioned in meta data, but not stored as own file  # line 299
        os.unlink("file2")  # line 300
        changes = sos.changes()  # line 301
        _.assertEqual(0, len(changes.additions))  # line 302
        _.assertEqual(1, len(changes.deletions))  # line 303
        _.assertEqual(1, len(changes.modifications))  # line 304
        sos.commit("modified")  # line 305
        _.assertEqual(1, len(os.listdir(branchFolder(0, 2))))  # no additional files, only mentions in metadata  # line 306
        try:  # expecting Exit due to no changes  # line 307
            sos.commit("nothing")  # expecting Exit due to no changes  # line 307
            _.fail()  # expecting Exit due to no changes  # line 307
        except:  # line 308
            pass  # line 308

    def testGetBranch(_):  # line 310
        m = sos.Metadata(os.getcwd())  # line 311
        m.branch = 1  # current branch  # line 312
        m.branches = {0: sos.BranchInfo(0, 0, "trunk")}  # line 313
        _.assertEqual(27, m.getBranchByName(27))  # line 314
        _.assertEqual(0, m.getBranchByName("trunk"))  # line 315
        _.assertEqual(1, m.getBranchByName(""))  # split from "/"  # line 316
        _.assertIsNone(m.getBranchByName("unknown"))  # line 317
        m.commits = {0: sos.CommitInfo(0, 0, "bla")}  # line 318
        _.assertEqual(13, m.getRevisionByName("13"))  # line 319
        _.assertEqual(0, m.getRevisionByName("bla"))  # line 320
        _.assertEqual(-1, m.getRevisionByName(""))  # split from "/"  # line 321

    def testTagging(_):  # line 323
        m = sos.Metadata(os.getcwd())  # line 324
        sos.offline()  # line 325
        _.createFile(111)  # line 326
        sos.commit("tag", ["--tag"])  # line 327
        _.createFile(2)  # line 328
        try:  # line 329
            sos.commit("tag")  # line 329
            _.fail()  # line 329
        except:  # line 330
            pass  # line 330
        sos.commit("tag-2", ["--tag"])  # line 331

    def testSwitch(_):  # line 333
        _.createFile(1, "x" * 100)  # line 334
        _.createFile(2, "y")  # line 335
        sos.offline("test")  # file1-2  in initial branch commit  # line 336
        sos.branch("second")  # file1-2  switch, having same files  # line 337
        sos.switch("0")  # no change  switch back, no problem  # line 338
        sos.switch("second")  # no change  # switch back, no problem  # line 339
        _.createFile(3, "y")  # generate a file  # line 340
        try:  # uncommited changes detected  # line 341
            sos.switch("test")  # uncommited changes detected  # line 341
            _.fail()  # uncommited changes detected  # line 341
        except SystemExit:  # line 342
            pass  # line 342
        sos.commit("Finish")  # file1-3  commit third file into branch second  # line 343
        sos.changes()  # line 344
        sos.switch("test")  # file1-2, remove file3 from file tree  # line 345
        _.assertFalse(_.existsFile(3))  # removed when switching back to test  # line 346
        out = wrapChannels(lambda: sos.status()).replace("\r", "")  # line 347
        _.assertIn("  * b00   'test'", out)  # line 348
        _.assertIn("    b01 'second'", out)  # line 349
        _.assertIn("(dirty)", out)  # one branch has commits  # line 350
        _.assertIn("(in sync)", out)  # the other doesn't  # line 351
        _.createFile(4, "xy")  # generate a file  # line 352
        sos.switch("second", "--force")  # avoids warning on uncommited changes, but keeps file4  # line 353
        _.assertFalse(_.existsFile(4))  # removed when forcedly switching back to test  # line 354
        _.assertTrue(_.existsFile(3))  # was restored from branch's revision r1  # line 355
        os.unlink("." + os.sep + "file1")  # remove old file1  # line 356
        sos.switch("test", "--force")  # should restore file1 and remove file3  # line 357
        _.assertTrue(_.existsFile(1))  # was restored from branch's revision r1  # line 358
        _.assertFalse(_.existsFile(3))  # was restored from branch's revision r1  # line 359

    def testAutoDetectVCS(_):  # line 361
        os.mkdir(".git")  # line 362
        sos.offline(sos.vcsBranches[sos.findSosVcsBase()[2]])  # create initial branch  # line 363
        with open(sos.metaFolder + os.sep + sos.metaFile, "r") as fd:  # line 364
            meta = fd.read()  # line 364
        _.assertTrue("\"master\"" in meta)  # line 365
        os.rmdir(".git")  # line 366

    def testUpdate(_):  # line 368
        sos.offline("trunk")  # create initial branch b0/r0  # line 369
        _.createFile(1, "x" * 100)  # line 370
        sos.commit("second")  # create b0/r1  # line 371

        sos.switch("/0")  # go back to b0/r0 - deletes file1  # line 373
        _.assertFalse(_.existsFile(1))  # line 374

        sos.update("/1")  # recreate file1  # line 376
        _.assertTrue(_.existsFile(1))  # line 377

        sos.commit("third", ["--force"])  # force because nothing to commit. should create r2 with same contents as r1, but as differential from r1, not from r0 (= no changes in meta folder)  # line 379
        _.assertTrue(os.path.exists(branchFolder(0, 2)))  # line 380
        _.assertTrue(os.path.exists(branchFolder(0, 2) + os.sep + sos.metaFile))  # line 381
        _.assertEqual(1, len(os.listdir(branchFolder(0, 2))))  # only meta data file, no differential files  # line 382

        sos.update("/1")  # do nothing, as nothing has changed  # line 384
        _.assertTrue(_.existsFile(1))  # line 385

        _.createFile(2, "y" * 100)  # line 387
#    out = wrapChannels(() -> sos.branch("other"))  # won't comply as there are changes
#    _.assertIn("--force", out)
        sos.branch("other", ["--force"])  # automatically including file 2 (as we are in simple mode)  # line 390
        _.assertTrue(_.existsFile(2))  # line 391
        sos.update("trunk", ["--add"])  # only add stuff  # line 392
        _.assertTrue(_.existsFile(2))  # line 393
        sos.update("trunk")  # nothing to do  # line 394
        _.assertFalse(_.existsFile(2))  # removes file not present in original branch  # line 395

        theirs = b"a\nb\nc\nd\ne\nf\ng\nh\nx\nx\nj\nk"  # line 397
        _.createFile(10, theirs)  # line 398
        mine = b"a\nc\nd\ne\ng\nf\nx\nh\ny\ny\nj"  # missing "b", inserted g, modified g->x, replace x/x -> y/y, removed k  # line 399
        _.createFile(11, mine)  # line 400
        _.assertEqual(b"a\nb\nc\nd\ne\nf\ng\nh\nx\nx\nj\nk", sos.merge(filename="." + os.sep + "file10", intoname="." + os.sep + "file11", mergeOperation=sos.MergeOperation.BOTH, conflictResolution=sos.ConflictResolution.THEIRS))  # completely recreated other file  # line 401
        _.assertEqual(b"a\nb\nc\nd\ne\ng\nf\ng\nx\nh\nx\nx\ny\ny\nj\nk", sos.merge(filename="." + os.sep + "file10", intoname="." + os.sep + "file11", mergeOperation=sos.MergeOperation.INSERT, conflictResolution=sos.ConflictResolution.MINE))  # line 402

    def testUpdate2(_):  # line 404
        _.createFile("test.txt", "x" * 10)  # line 405
        sos.offline("trunk", ["--strict"])  # use strict mode, as timestamp differences are too small for testing  # line 406
        sos.branch("mod")  # line 407
        _.createFile("test.txt", "x" * 5 + "y" * 5)  # line 408
        time.sleep(FS_PRECISION)  # line 409
        sos.commit("mod")  # create b0/r1  # line 410
        sos.switch("trunk", ["--force"])  # should replace contents, force in case some other files were modified (e.g. during working on the code) TODO investigate more  # line 411
        with open("test.txt", "rb") as fd:  # line 412
            _.assertEqual(b"x" * 10, fd.read())  # line 412
        sos.update("mod", ["--theirs"])  # integrate changes TODO same with ask -> theirs  # line 413
        with open("test.txt", "rb") as fd:  # line 414
            _.assertEqual(b"x" * 5 + b"y" * 5, fd.read())  # line 414
        _.createFile("test.txt", "x" * 10)  # line 415
        mockInput(["t"], lambda: sos.update("mod", ["--ask"]))  # same as above with interaction -> theirs  # line 416
        with open("test.txt", "rb") as fd:  # line 417
            _.assertEqual(b"x" * 5 + b"y" * 5, fd.read())  # line 417

    def testIsTextType(_):  # line 419
        m = sos.Metadata(".")  # line 420
        m.c.texttype = ["*.x", "*.md", "*.md.*"]  # line 421
        m.c.bintype = ["*.md.confluence"]  # line 422
        _.assertTrue(m.isTextType("ab.txt"))  # line 423
        _.assertTrue(m.isTextType("./ab.txt"))  # line 424
        _.assertTrue(m.isTextType("bc/ab.txt"))  # line 425
        _.assertFalse(m.isTextType("bc/ab."))  # line 426
        _.assertTrue(m.isTextType("23_3.x.x"))  # line 427
        _.assertTrue(m.isTextType("dfg/dfglkjdf7/test.md"))  # line 428
        _.assertTrue(m.isTextType("./test.md.pdf"))  # line 429
        _.assertFalse(m.isTextType("./test_a.md.confluence"))  # line 430

    def testEolDet(_):  # line 432
        ''' Check correct end-of-line detection. '''  # line 433
        _.assertEqual(b"\n", sos.eoldet(b"a\nb"))  # line 434
        _.assertEqual(b"\r\n", sos.eoldet(b"a\r\nb\r\n"))  # line 435
        _.assertEqual(b"\r", sos.eoldet(b"\ra\rb"))  # line 436
        _.assertAllIn(["Inconsistent", "with "], wrapChannels(lambda: _.assertEqual(b"\n", sos.eoldet(b"\r\na\r\nb\n"))))  # line 437
        _.assertAllIn(["Inconsistent", "without"], wrapChannels(lambda: _.assertEqual(b"\n", sos.eoldet(b"\ra\nnb\n"))))  # line 438
        _.assertIsNone(sos.eoldet(b""))  # line 439
        _.assertIsNone(sos.eoldet(b"sdf"))  # line 440

    def testMerge(_):  # line 442
        ''' Check merge results depending on conflict solution options. '''  # line 443
        a = b"a\nb\ncc\nd"  # line 444
        b = b"a\nb\nee\nd"  # line 445
        _.assertEqual(b"a\nb\ncc\nee\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.INSERT))  # means insert changes from a into b, but don't replace  # line 446
        _.assertEqual(b"a\nb\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.REMOVE))  # line 447
        _.assertEqual(b"a\nb\ncc\nd", sos.merge(a, b, mergeOperation=sos.MergeOperation.BOTH))  # line 448
# Now test intra-line merging without conflicts
        _.assertEqual(b"a\nbc d\ne", sos.merge(b"a\nbc d\ne", b"a\nbcd\ne", mergeOperation=sos.MergeOperation.INSERT))  # because it's a deletion ['  a', '- bc d', '?   -\n', '+ bcd', '  e']  # line 450
        _.assertEqual(b"a\nbcd\ne", sos.merge(b"a\nbc d\ne", b"a\nbcd\ne", mergeOperation=sos.MergeOperation.REMOVE))  # ['  a', '- bc d', '?   -\n', '+ bcd', '  e']  # line 451
        _.assertEqual(b"a\nbc d\ne", sos.merge(b"a\nbcd\ne", b"a\nbc d\ne", mergeOperation=sos.MergeOperation.INSERT))  # nothing to insert  # line 452
        _.assertEqual(b"a\nbcd\ne", sos.merge(b"a\nbcd\ne", b"a\nbc d\ne", mergeOperation=sos.MergeOperation.REMOVE))  # remove space  # line 453
# Test with change + insert (conflict)
        _.assertEqual(b"a\nb fdd d\ne", sos.merge(b"a\nb cd d\ne", b"a\nb fdd d\ne", conflictResolution=sos.ConflictResolution.MINE))  # line 455
        _.assertEqual(b"a\nb cd d\ne", sos.merge(b"a\nb cd d\ne", b"a\nb fdd d\ne", conflictResolution=sos.ConflictResolution.THEIRS))  # ['  a', '- b cd d', '?   ^\n', '+ b fdd d', '?   ^^\n', '  e']  # line 456
        _.assertEqual(b"a\nb fdd d\ne", mockInput(["i"], lambda: sos.merge(b"a\nb cd d\ne", b"a\nb fdd d\ne", conflictResolution=sos.ConflictResolution.ASK)))  # ['  a', '- b cd d', '?   ^\n', '+ b fdd d', '?   ^^\n', '  e']  # line 457
        _.assertEqual(b"a\nb cd d\ne", mockInput(["t"], lambda: sos.merge(b"a\nb cd d\ne", b"a\nb fdd d\ne", conflictResolution=sos.ConflictResolution.ASK)))  # ['  a', '- b cd d', '?   ^\n', '+ b fdd d', '?   ^^\n', '  e']  # line 458
        _.assertEqual(b"abbc", sos.merge(b"abbc", b"addc", mergeOperation=sos.MergeOperation.BOTH, conflictResolution=sos.ConflictResolution.THEIRS))  # line 459
        _.assertEqual(b"a\nbb\nc", sos.merge(b"a\nbb\nc", b"a\ndd\nc", mergeOperation=sos.MergeOperation.BOTH, conflictResolution=sos.ConflictResolution.THEIRS))  # line 460
        _.assertIn("Differing EOL-styles", wrapChannels(lambda: sos.merge(b"a\nb", b"a\r\nb")))  # expect warning  # line 461
        _.assertIn(b"a\r\nb", sos.merge(b"a\nb", b"a\r\nb"))  # in doubt, use "mine" CR-LF  # line 462

    def testPickyMode(_):  # line 464
        ''' Confirm that picky mode reset tracked patterns after commits. '''  # line 465
        sos.offline("trunk", ["--picky"])  # line 466
        sos.add(".", "./file?", ["--force"])  # line 467
        _.createFile(1, "aa")  # line 468
        sos.commit("First")  # add one file  # line 469
        _.assertEqual(2, len(os.listdir(branchFolder(0, 1))))  # line 470
        _.createFile(2, "b")  # line 471
        try:  # add nothing, because picky  # line 472
            sos.commit("Second")  # add nothing, because picky  # line 472
        except:  # line 473
            pass  # line 473
        sos.add(".", "./file?")  # line 474
        sos.commit("Third")  # line 475
        _.assertEqual(2, len(os.listdir(branchFolder(0, 2))))  # line 476
        out = wrapChannels(lambda: sos.log([])).replace("\r", "")  # line 477
        _.assertIn("  * r2", out)  # line 478

    def testTrackedSubfolder(_):  # line 480
        ''' See if patterns for files in sub folders are picked up correctly. '''  # line 481
        os.mkdir("." + os.sep + "sub")  # line 482
        sos.offline("trunk", ["--track"])  # line 483
        _.createFile(1, "x")  # line 484
        _.createFile(1, "x", prefix="sub")  # line 485
        sos.add(".", "./file?")  # add glob pattern to track  # line 486
        sos.commit("First")  # line 487
        _.assertEqual(2, len(os.listdir(branchFolder(0, 1))))  # one new file + meta file  # line 488
        sos.add(".", "sub/file?")  # add glob pattern to track  # line 489
        sos.commit("Second")  # one new file + meta  # line 490
        _.assertEqual(2, len(os.listdir(branchFolder(0, 1))))  # one new file + meta file  # line 491
        os.unlink("file1")  # remove from basefolder  # line 492
        _.createFile(2, "y")  # line 493
        sos.rm(".", "sub/file?")  # line 494
        try:  # raises Exit. TODO test the "suggest a pattern" case  # line 495
            sos.rm(".", "sub/bla")  # raises Exit. TODO test the "suggest a pattern" case  # line 495
            _.fail()  # raises Exit. TODO test the "suggest a pattern" case  # line 495
        except:  # line 496
            pass  # line 496
        sos.commit("Third")  # line 497
        _.assertEqual(2, len(os.listdir(branchFolder(0, 2))))  # one new file + meta  # line 498
# TODO check if /file1 and sub/file1 were removed from index

    def testTrackedMode(_):  # line 501
        ''' Difference in semantics vs simple mode:
          - For remote/other branch we can only know and consider tracked files, thus ignoring all complexity stemming from handling addition of untracked files.
          - For current branch, we can take into account tracked and untracked ones, in theory, but it doesn't make sense.
        In conclusion, using the union of tracking patterns from both sides to find affected files makes sense, but disallow deleting files not present in remote branch.
    '''  # line 506
        sos.offline("test", options=["--track"])  # set up repo in tracking mode (SVN- or gitless-style)  # line 507
        _.createFile(1)  # line 508
        _.createFile("a123a")  # untracked file "a123a"  # line 509
        sos.add(".", "./file?")  # add glob tracking pattern  # line 510
        sos.commit("second")  # versions "file1"  # line 511
        _.assertEqual(2, len(os.listdir(branchFolder(0, 1))))  # one new file + meta file  # line 512
        out = wrapChannels(lambda: sos.status()).replace("\r", "")  # line 513
        _.assertIn("  | ./file?", out)  # line 514

        _.createFile(2)  # untracked file "file2"  # line 516
        sos.commit("third")  # versions "file2"  # line 517
        _.assertEqual(2, len(os.listdir(branchFolder(0, 2))))  # one new file + meta file  # line 518

        os.mkdir("." + os.sep + "sub")  # line 520
        _.createFile(3, prefix="sub")  # untracked file "sub/file3"  # line 521
        sos.commit("fourth", ["--force"])  # no tracking pattern matches the subfolder  # line 522
        _.assertEqual(1, len(os.listdir(branchFolder(0, 3))))  # meta file only, no other tracked path/file  # line 523

        sos.branch("Other")  # second branch containing file1 and file2 tracked by "./file?"  # line 525
        sos.rm(".", "./file?")  # remove tracking pattern, but don't touch previously created and versioned files  # line 526
        sos.add(".", "./a*a")  # add tracking pattern  # line 527
        changes = sos.changes()  # should pick up addition only, because tracked, but not the deletion, as not tracked anymore  # line 528
        _.assertEqual(0, len(changes.modifications))  # line 529
        _.assertEqual(0, len(changes.deletions))  # not tracked anymore, but contained in version history and not removed  # line 530
        _.assertEqual(1, len(changes.additions))  # detected one addition "a123a", but won't recognize untracking files as deletion  # line 531

        sos.commit("Second_2")  # line 533
        _.assertEqual(2, len(os.listdir(branchFolder(1, 1))))  # "a123a" + meta file  # line 534
        _.assertTrue(os.path.exists("." + os.sep + "file1"))  # line 535
        _.assertTrue(os.path.exists("." + os.sep + "file2"))  # line 536

        sos.switch("test")  # go back to first branch - tracks only "file?", but not "a*a"  # line 538
        _.assertTrue(os.path.exists("." + os.sep + "file1"))  # line 539
        _.assertTrue(os.path.exists("." + os.sep + "a123a"))  # should not have been touched so far  # line 540

        sos.update("Other")  # integrate tracked files and tracking pattern from second branch into working state of master branch  # line 542
        _.assertTrue(os.path.exists("." + os.sep + "file1"))  # line 543
        _.assertTrue(os.path.exists("." + os.sep + "a123a"))  # line 544

        _.createFile("axxxa")  # new file that should be tracked on "test" now that we integrated "Other"  # line 546
        sos.commit("fifth")  # create new revision after integrating updates from second branch  # line 547
        _.assertEqual(3, len(os.listdir(branchFolder(0, 4))))  # one new file from other branch + one new in current folder + meta file  # line 548
        sos.switch("Other")  # switch back to just integrated branch that tracks only "a*a" - shouldn't do anything  # line 549
        _.assertTrue(os.path.exists("." + os.sep + "file1"))  # line 550
        _.assertTrue(os.path.exists("." + os.sep + "a123a"))  # line 551
        _.assertFalse(os.path.exists("." + os.sep + "axxxa"))  # because tracked in both branches, but not present in other -> delete in file tree TODO document  # line 552
# TODO test switch --meta

    def testLsTracked(_):  # line 555
        sos.offline("test", options=["--track"])  # set up repo in tracking mode (SVN- or gitless-style)  # line 556
        _.createFile(1)  # line 557
        _.createFile("foo")  # line 558
        sos.add(".", "./file*")  # capture one file  # line 559
        out = sos.safeSplit(wrapChannels(lambda: sos.ls()).replace("\r", ""), "\n")  # line 560
        _.assertInAny('TRK file1 by "./file*"', out)  # line 561
        _.assertNotInAny('    file1 by "./file*"', out)  # line 562
        _.assertInAny("    foo", out)  # line 563

    def testCompression(_):  # line 565
        _.createFile(1)  # line 566
        sos.offline("master", options=["--plain", "--force"])  # line 567
        _.assertTrue(_.existsFile(branchFolder(0, 0) + os.sep + "b9ee10a87f612e299a6eb208210bc0898092a64c48091327cc2aaeee9b764ffa", b"x" * 10))  # line 568
        setRepoFlag("compress", True)  # was plain = uncompressed before  # line 569
        _.createFile(2)  # line 570
        sos.commit("Added file2")  # line 571
        _.assertTrue(_.existsFile(branchFolder(0, 1) + os.sep + "03b69bc801ae11f1ff2a71a50f165996d0ad681b4f822df13329a27e53f0fcd2"))  # exists  # line 572
        _.assertFalse(_.existsFile(branchFolder(0, 1) + os.sep + "03b69bc801ae11f1ff2a71a50f165996d0ad681b4f822df13329a27e53f0fcd2", b"x" * 10))  # but is compressed instead  # line 573

    def testConfigVariations(_):  # line 575
        def makeRepo():  # line 576
            try:  # line 577
                os.unlink("file1")  # line 577
            except:  # line 578
                pass  # line 578
            sos.offline("master", options=["--plain", "--force"])  # line 579
            _.createFile(1)  # line 580
            sos.commit("Added file1")  # line 581
        sos.config("set", ["strict", "on"])  # line 582
        makeRepo()  # line 583
        _.assertTrue(checkRepoFlag("strict", True))  # line 584
        sos.config("set", ["strict", "off"])  # line 585
        makeRepo()  # line 586
        _.assertTrue(checkRepoFlag("strict", False))  # line 587
        sos.config("set", ["strict", "yes"])  # line 588
        makeRepo()  # line 589
        _.assertTrue(checkRepoFlag("strict", True))  # line 590
        sos.config("set", ["strict", "no"])  # line 591
        makeRepo()  # line 592
        _.assertTrue(checkRepoFlag("strict", False))  # line 593
        sos.config("set", ["strict", "1"])  # line 594
        makeRepo()  # line 595
        _.assertTrue(checkRepoFlag("strict", True))  # line 596
        sos.config("set", ["strict", "0"])  # line 597
        makeRepo()  # line 598
        _.assertTrue(checkRepoFlag("strict", False))  # line 599
        sos.config("set", ["strict", "true"])  # line 600
        makeRepo()  # line 601
        _.assertTrue(checkRepoFlag("strict", True))  # line 602
        sos.config("set", ["strict", "false"])  # line 603
        makeRepo()  # line 604
        _.assertTrue(checkRepoFlag("strict", False))  # line 605
        sos.config("set", ["strict", "enable"])  # line 606
        makeRepo()  # line 607
        _.assertTrue(checkRepoFlag("strict", True))  # line 608
        sos.config("set", ["strict", "disable"])  # line 609
        makeRepo()  # line 610
        _.assertTrue(checkRepoFlag("strict", False))  # line 611
        sos.config("set", ["strict", "enabled"])  # line 612
        makeRepo()  # line 613
        _.assertTrue(checkRepoFlag("strict", True))  # line 614
        sos.config("set", ["strict", "disabled"])  # line 615
        makeRepo()  # line 616
        _.assertTrue(checkRepoFlag("strict", False))  # line 617
        try:  # line 618
            sos.config("set", ["strict", "nope"])  # line 618
            _.fail()  # line 618
        except:  # line 619
            pass  # line 619

    def testLsSimple(_):  # line 621
        _.createFile(1)  # line 622
        _.createFile("foo")  # line 623
        _.createFile("ign1")  # line 624
        _.createFile("ign2")  # line 625
        sos.offline("test")  # set up repo in tracking mode (SVN- or gitless-style)  # line 626
        sos.config("set", ["ignores", "ign1"])  # define an ignore pattern  # line 627
        sos.config("add", ["ignores", "ign2"])  # define an ignore pattern  # line 628
        sos.config("set", ["ignoresWhitelist", "ign1;ign2"])  # define a list of ignore patterns  # line 629
        out = wrapChannels(lambda: sos.config("show")).replace("\r", "")  # line 630
        _.assertIn("ignores => ['ign1', 'ign2']", out)  # line 631
        out = sos.safeSplit(wrapChannels(lambda: sos.ls()).replace("\r", ""), "\n")  # line 632
        _.assertInAny('    file1', out)  # line 633
        _.assertInAny('    ign1', out)  # line 634
        _.assertInAny('    ign2', out)  # line 635
        try:  # line 636
            sos.config("rm", ["foo", "bar"])  # line 636
            _.fail()  # line 636
        except:  # line 637
            pass  # line 637
        try:  # line 638
            sos.config("rm", ["ignores", "foo"])  # line 638
            _.fail()  # line 638
        except:  # line 639
            pass  # line 639
        sos.config("rm", ["ignores", "ign1"])  # line 640
        sos.config("unset", ["ignoresWhitelist"])  # remove ignore pattern  # line 641
        out = sos.safeSplit(wrapChannels(lambda: sos.ls()).replace("\r", ""), "\n")  # line 642
        _.assertInAny('    ign1', out)  # line 643
        _.assertInAny('IGN ign2', out)  # line 644
        _.assertNotInAny('    ign2', out)  # line 645

    def testWhitelist(_):  # line 647
# TODO test same for simple mode
        _.createFile(1)  # line 649
        sos.defaults.ignores[:] = ["file*"]  # replace in-place  # line 650
        sos.offline("xx", ["--track", "--strict"])  # because nothing to commit due to ignore pattern  # line 651
        sos.add(".", "./file*")  # add tracking pattern for "file1"  # line 652
        sos.commit(options=["--force"])  # attempt to commit the file  # line 653
        _.assertEqual(1, len(os.listdir(branchFolder(0, 1))))  # only meta data, file1 was ignored  # line 654
        try:  # Exit because dirty  # line 655
            sos.online()  # Exit because dirty  # line 655
            _.fail()  # Exit because dirty  # line 655
        except:  # exception expected  # line 656
            pass  # exception expected  # line 656
        _.createFile("x2")  # add another change  # line 657
        sos.add(".", "./x?")  # add tracking pattern for "file1"  # line 658
        try:  # force beyond dirty flag check  # line 659
            sos.online(["--force"])  # force beyond dirty flag check  # line 659
            _.fail()  # force beyond dirty flag check  # line 659
        except:  # line 660
            pass  # line 660
        sos.online(["--force", "--force"])  # force beyond file tree modifications check  # line 661
        _.assertFalse(os.path.exists(sos.metaFolder))  # line 662

        _.createFile(1)  # line 664
        sos.defaults.ignoresWhitelist[:] = ["file*"]  # line 665
        sos.offline("xx", ["--track"])  # line 666
        sos.add(".", "./file*")  # line 667
        sos.commit()  # should NOT ask for force here  # line 668
        _.assertEqual(2, len(os.listdir(branchFolder(0, 1))))  # meta data and "file1", file1 was whitelisted  # line 669

    def testRemove(_):  # line 671
        _.createFile(1, "x" * 100)  # line 672
        sos.offline("trunk")  # line 673
        try:  # line 674
            sos.delete("trunk")  # line 674
            _fail()  # line 674
        except:  # line 675
            pass  # line 675
        _.createFile(2, "y" * 10)  # line 676
        sos.branch("added")  # line 677
        sos.delete("trunk")  # line 678
        _.assertEqual(2, len(os.listdir("." + os.sep + sos.metaFolder)))  # meta data file and "b1"  # line 679
        _.assertTrue(os.path.exists("." + os.sep + sos.metaFolder + os.sep + "b1"))  # line 680
        _.assertFalse(os.path.exists("." + os.sep + sos.metaFolder + os.sep + "b0"))  # line 681
        sos.branch("next")  # line 682
        _.createFile(3, "y" * 10)  # make a change  # line 683
        sos.delete("added", "--force")  # should succeed  # line 684

    def testUsage(_):  # line 686
        sos.usage()  # line 687

    def testOnly(_):  # line 689
        _.assertEqual((_coconut.frozenset(((".", "./A"), ("x", "x/B"))), _coconut.frozenset(((".", "./C"),))), sos.parseOnlyOptions(".", ["abc", "def", "--only", "A", "--x", "--only", "x/B", "--except", "C", "--only"]))  # line 690
        _.assertEqual(_coconut.frozenset(("B",)), sos.conditionalIntersection(_coconut.frozenset(("A", "B", "C")), _coconut.frozenset(("B", "D"))))  # line 691
        _.assertEqual(_coconut.frozenset(("B", "D")), sos.conditionalIntersection(_coconut.frozenset(), _coconut.frozenset(("B", "D"))))  # line 692
        _.assertEqual(_coconut.frozenset(("B", "D")), sos.conditionalIntersection(None, _coconut.frozenset(("B", "D"))))  # line 693

    def testDiff(_):  # line 695
        sos.offline(options=["--strict"])  # line 696
        _.createFile(1)  # line 697
        _.createFile(2)  # line 698
        sos.commit()  # line 699
        _.createFile(1, "sdfsdgfsdf")  # line 700
        _.createFile(2, "12343")  # line 701
        sos.commit()  # line 702
        _.createFile(1, "foobar")  # line 703
        _.assertAllIn(["MOD ./file1", "MOD ./file2", "DIF ./file1", "- | 0000 |xxxxxxxxxx|", "+ | 0000 |foobar|"], wrapChannels(lambda: sos.diff("/-2")))  # vs. second last  # line 704
        _.assertNotIn("MOD ./file1", wrapChannels(lambda: sos.diff("/-2", onlys=_coconut.frozenset(("./file2",)))))  # line 705

    def testFindBase(_):  # line 707
        old = os.getcwd()  # line 708
        try:  # line 709
            os.mkdir("." + os.sep + ".git")  # line 710
            os.makedirs("." + os.sep + "a" + os.sep + sos.metaFolder)  # line 711
            os.makedirs("." + os.sep + "a" + os.sep + "b")  # line 712
            os.chdir("a" + os.sep + "b")  # line 713
            s, vcs, cmd = sos.findSosVcsBase()  # line 714
            _.assertIsNotNone(s)  # line 715
            _.assertIsNotNone(vcs)  # line 716
            _.assertEqual("git", cmd)  # line 717
        finally:  # line 718
            os.chdir(old)  # line 718


if __name__ == '__main__':  # line 721
    logging.basicConfig(level=logging.DEBUG if '-v' in sys.argv or os.getenv("DEBUG", "false").strip().lower() == "true" or os.getenv("CI", "false").strip().lower() == "true" else logging.INFO, stream=sys.stderr, format="%(asctime)-23s %(levelname)-8s %(name)s:%(lineno)d | %(message)s" if '-v' in sys.argv or os.getenv("DEBUG", "false") == "true" else "%(message)s")  # line 722
    if configr:  # line 723
        c = configr.Configr("sos")  # line 724
        c.loadSettings()  # line 724
        if len(c.keys()) > 0:  # line 725
            sos.Exit("Cannot run test suite with existing local SOS user configuration (would affect results)")  # line 725
    unittest.main(testRunner=debugTestRunner() if '-v' in sys.argv and not os.getenv("CI", "false").lower() == "true" else None)  # warnings = "ignore")  # line 726



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x8a841ccc

# Compiled with Coconut version 1.3.1 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_tail_call, _coconut_tco, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

# Utiliy functions
import bz2  # line 2
import codecs  # line 2
import difflib  # line 2
import hashlib  # line 2
import logging  # line 2
import os  # line 2
sys = _coconut_sys  # line 2
import time  # line 2
try:  # line 3
    from typing import Any  # only required for mypy  # line 4
    from typing import Callable  # only required for mypy  # line 4
    from typing import Dict  # only required for mypy  # line 4
    from typing import IO  # only required for mypy  # line 4
    from typing import List  # only required for mypy  # line 4
    from typing import Tuple  # only required for mypy  # line 4
    from typing import Type  # only required for mypy  # line 4
    from typing import TypeVar  # only required for mypy  # line 4
    from typing import Union  # only required for mypy  # line 4
    Number = Union[int, float]  # line 5
except:  # typing not available (e.g. Python 2)  # line 6
    pass  # typing not available (e.g. Python 2)  # line 6
try:  # line 7
    import wcwidth  # line 7
except:  # optional dependency  # line 8
    pass  # optional dependency  # line 8
longint = eval("long") if sys.version_info.major < 3 else int  # type: Type  # for Python 2 compatibility  # line 9


# Classes
class CallOnce:  # line 13
    ''' This object can be called any time, but will call the wrapped function only once. '''  # line 14
    def __init__(_, func: 'Callable[[], None]') -> 'None':  # line 15
        _.func = func  # type: _coconut.typing.Optional[Callable[[], None]]  # line 15
    def __call__(_) -> 'None':  # sets to None  # line 16
        _.func = _.func() if _.func else None  # sets to None  # line 16

class Accessor(dict):  # line 18
    ''' Dictionary with attribute access. Writing only supported via dictionary access. '''  # line 19
    def __init__(_, mapping: 'Dict[str, Any]') -> 'None':  # line 20
        dict.__init__(_, mapping)  # line 20
    @_coconut_tco  # line 21
    def __getattribute__(_, name: 'str') -> 'Any':  # line 21
        try:  # line 22
            return _[name]  # line 22
        except:  # line 23
            return _coconut_tail_call(dict.__getattribute__, _, name)  # line 23

class Counter:  # line 25
    ''' A simple counter. Can be augmented to return the last value instead. '''  # line 26
    def __init__(_, initial: 'Number'=0) -> 'None':  # line 27
        _.value = initial  # type: Number  # line 27
    def inc(_, by=1) -> 'Number':  # line 28
        _.value += by  # line 28
        return _.value  # line 28

class Logger:  # line 30
    ''' Logger that supports many items. '''  # line 31
    def __init__(_, log):  # line 32
        _._log = log  # line 32
    def debug(_, *s):  # line 33
        _._log.debug(sjoin(*s))  # line 33
    def info(_, *s):  # line 34
        _._log.info(sjoin(*s))  # line 34
    def warn(_, *s):  # line 35
        _._log.warning(sjoin(*s))  # line 35
    def error(_, *s):  # line 36
        _._log.error(sjoin(*s))  # line 36


# Constants
_log = Logger(logging.getLogger(__name__))  # line 40
debug, info, warn, error = _log.debug, _log.info, _log.warn, _log.error  # line 40
UTF8 = "utf_8"  # early used constant, not defined in standard library  # line 41
CONFIGURABLE_FLAGS = ["strict", "track", "picky", "compress"]  # line 42
CONFIGURABLE_LISTS = ["texttype", "bintype", "ignores", "ignoreDirs", "ignoresWhitelist", "ignoreDirsWhitelist"]  # line 43
TRUTH_VALUES = ["true", "yes", "on", "1", "enable", "enabled"]  # all lower-case normalized  # line 44
FALSE_VALUES = ["false", "no", "off", "0", "disable", "disabled"]  # line 45
PROGRESS_MARKER = ["|", "/", "-", "\\"]  # line 46
metaFolder = ".sos"  # type: str  # line 47
metaFile = ".meta"  # type: str  # line 48
bufSize = 1 << 20  # type: int  # 1 MiB  # line 49
SVN = "svn"  # line 50
SLASH = "/"  # line 51
vcsFolders = {".svn": SVN, ".git": "git", ".bzr": "bzr", ".hg": "hg", ".fslckout": "fossil", "_FOSSIL_": "fossil", ".CVS": "cvs"}  # type: Dict[str, str]  # line 52
vcsBranches = {SVN: "trunk", "git": "master", "bzr": "trunk", "hg": "default", "fossil": None, "cvs": None}  # type: Dict[str, _coconut.typing.Optional[str]]  # line 53


# Value types
class BranchInfo(_coconut_NamedTuple("BranchInfo", [("number", 'int'), ("ctime", 'int'), ("name", '_coconut.typing.Optional[str]'), ("inSync", 'bool'), ("tracked", 'List[str]')])):  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 57
    __slots__ = ()  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 57
    __ne__ = _coconut.object.__ne__  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 57
    def __new__(_cls, number, ctime, name=None, inSync=False, tracked=[]):  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 57
        return _coconut.tuple.__new__(_cls, (number, ctime, name, inSync, tracked))  # tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place  # line 57
# tracked is a list on purpose, as serialization to JSON needs effort and frequent access is not taking place
class CommitInfo(_coconut_NamedTuple("CommitInfo", [("number", 'int'), ("ctime", 'int'), ("message", '_coconut.typing.Optional[str]')])):  # line 58
    __slots__ = ()  # line 58
    __ne__ = _coconut.object.__ne__  # line 58
    def __new__(_cls, number, ctime, message=None):  # line 58
        return _coconut.tuple.__new__(_cls, (number, ctime, message))  # line 58

class PathInfo(_coconut_NamedTuple("PathInfo", [("nameHash", 'str'), ("size", '_coconut.typing.Optional[int]'), ("mtime", 'int'), ("hash", '_coconut.typing.Optional[str]')])):  # size == None means deleted in this revision  # line 59
    __slots__ = ()  # size == None means deleted in this revision  # line 59
    __ne__ = _coconut.object.__ne__  # size == None means deleted in this revision  # line 59
class ChangeSet(_coconut_NamedTuple("ChangeSet", [("additions", 'Dict[str, PathInfo]'), ("deletions", 'Dict[str, PathInfo]'), ("modifications", 'Dict[str, PathInfo]')])):  # avoid default assignment of {} as it leads to runtime errors (contains data on init for unknown reason)  # line 60
    __slots__ = ()  # avoid default assignment of {} as it leads to runtime errors (contains data on init for unknown reason)  # line 60
    __ne__ = _coconut.object.__ne__  # avoid default assignment of {} as it leads to runtime errors (contains data on init for unknown reason)  # line 60
class Range(_coconut_NamedTuple("Range", [("tipe", 'int'), ("indexes", '_coconut.typing.Sequence[int]')])):  # MergeBlockType[1,2,4], line number, length  # line 61
    __slots__ = ()  # MergeBlockType[1,2,4], line number, length  # line 61
    __ne__ = _coconut.object.__ne__  # MergeBlockType[1,2,4], line number, length  # line 61
class MergeBlock(_coconut_NamedTuple("MergeBlock", [("tipe", 'int'), ("lines", '_coconut.typing.Sequence[str]'), ("line", 'int'), ("replaces", '_coconut.typing.Optional[MergeBlock]'), ("changes", '_coconut.typing.Optional[Range]')])):  # line 62
    __slots__ = ()  # line 62
    __ne__ = _coconut.object.__ne__  # line 62
    def __new__(_cls, tipe, lines, line, replaces=None, changes=None):  # line 62
        return _coconut.tuple.__new__(_cls, (tipe, lines, line, replaces, changes))  # line 62



# Enums
class ConflictResolution:  # prefer their changes, prefer my changes, ask user for each change, go to next deeper level (e.g. from files to lines, characters)  # line 66
    THEIRS, MINE, ASK, NEXT = range(4)  # prefer their changes, prefer my changes, ask user for each change, go to next deeper level (e.g. from files to lines, characters)  # line 66
class MergeOperation:  # insert remote changes into current, remove remote deletions from current, do both (replicate remote state) TODO handle inline-operations separate?  # line 67
    INSERT, REMOVE, BOTH = 1, 2, 3  # insert remote changes into current, remove remote deletions from current, do both (replicate remote state) TODO handle inline-operations separate?  # line 67
class MergeBlockType:  # modify = intra-line changes  # line 68
    KEEP, INSERT, REMOVE, REPLACE, MODIFY, MOVE = range(6)  # modify = intra-line changes  # line 68


# Functions
try:  # line 72
    import chardet  # https://github.com/chardet/chardet  # line 73
    def detectEncoding(binary: 'bytes') -> 'str':  # line 74
        return chardet.detect(binary)["encoding"]  # line 74
except:  # line 75
    def detectEncoding(binary: 'bytes') -> 'str':  # Guess the encoding  # line 76
        ''' Fallback if chardet library missing. '''  # line 77
        try:  # line 78
            binary.decode(UTF8)  # line 78
            return UTF8  # line 78
        except UnicodeError:  # line 79
            pass  # line 79
        try:  # line 80
            binary.decode("utf_16")  # line 80
            return "utf_16"  # line 80
        except UnicodeError:  # line 81
            pass  # line 81
        try:  # line 82
            binary.decode("cp1252")  # line 82
            return "cp1252"  # line 82
        except UnicodeError:  # line 83
            pass  # line 83
        return "ascii"  # this code will never be reached, as above is an 8-bit charset that always matches  # line 84

@_coconut_tco  # line 86
def wcswidth(string: 'str') -> 'int':  # line 86
    l = None  # type: int  # line 87
    try:  # line 88
        l = wcwidth.wcswitdh(string)  # line 89
        if l < 0:  # line 90
            return len(string)  # line 90
    except:  # line 91
        return _coconut_tail_call(len, string)  # line 91
    return l  # line 92

def openIt(file: 'str', mode: 'str', compress: 'bool'=False) -> 'IO':  # Abstraction for opening both compressed and plain files  # line 94
    return bz2.BZ2File(file, mode) if compress else open(file, mode + "b")  # Abstraction for opening both compressed and plain files  # line 94

def eoldet(file: 'bytes') -> '_coconut.typing.Optional[bytes]':  # line 96
    ''' Determine EOL style from a binary string. '''  # line 97
    lf = file.count(b"\n")  # type: int  # line 98
    cr = file.count(b"\r")  # type: int  # line 99
    crlf = file.count(b"\r\n")  # type: int  # line 100
    if crlf > 0:  # DOS/Windows/Symbian etc.  # line 101
        if lf != crlf or cr != crlf:  # line 102
            warn("Inconsistent CR/NL count with CR+NL. Mixed EOL style detected, may cause problems during merge")  # line 102
        return b"\r\n"  # line 103
    if lf != 0 and cr != 0:  # line 104
        warn("Inconsistent CR/NL count without CR+NL. Mixed EOL style detected, may cause problems during merge")  # line 104
    if lf > cr:  # Linux/Unix  # line 105
        return b"\n"  # Linux/Unix  # line 105
    if cr > lf:  # older 8-bit machines  # line 106
        return b"\r"  # older 8-bit machines  # line 106
    return None  # no new line contained, cannot determine  # line 107

def Exit(message: 'str'="") -> 'None':  # line 109
    print("[EXIT] " + message + ".", file=sys.stderr) if str != "" else None  # line 109
    sys.exit(1)  # line 109

@_coconut_tco  # referenceso __builtin__.raw_input on Python 2  # line 111
def user_input(msg: 'str') -> 'str':  # referenceso __builtin__.raw_input on Python 2  # line 111
    return _coconut_tail_call(input, msg)  # referenceso __builtin__.raw_input on Python 2  # line 111

try:  # line 113
    Splittable = TypeVar("Splittable", str, bytes)  # line 113
except:  # Python 2  # line 114
    pass  # Python 2  # line 114
def safeSplit(s: 'Splittable', d: '_coconut.typing.Optional[Splittable]'=None) -> '_coconut.typing.Sequence[Splittable]':  # line 115
    return s.split(("\n" if isinstance(s, str) else b"\n") if d is None else d) if len(s) > 0 else []  # line 115

def ajoin(sep: 'str', seq: '_coconut.typing.Sequence[str]', nl="") -> 'str':  # line 117
    return sep + (nl + sep).join(seq) if seq else ""  # line 117

@_coconut_tco  # line 119
def sjoin(*s: 'Tuple[Any]') -> 'str':  # line 119
    return _coconut_tail_call(" ".join, [str(e) for e in s if e != ''])  # line 119

@_coconut_tco  # line 121
def hashStr(datas: 'str') -> 'str':  # line 121
    return _coconut_tail_call(hashlib.sha256(datas.encode(UTF8)).hexdigest)  # line 121

def modified(changes: 'ChangeSet', onlyBinary: 'bool'=False) -> 'bool':  # line 123
    return len(changes.additions) > 0 or len(changes.deletions) > 0 or len(changes.modifications) > 0  # line 123

def getTermWidth() -> 'int':  # line 125
    try:  # line 125
        import termwidth  # line 126
    except:  # line 127
        return 80  # line 127
    return termwidth.getTermWidth()[0]  # line 128

def hashFile(path: 'str', compress: 'bool', saveTo: '_coconut.typing.Optional[str]'=None) -> 'Tuple[str, longint]':  # line 130
    ''' Calculate hash of file contents, and return compressed sized, if in write mode, or zero. '''  # line 131
    _hash = hashlib.sha256()  # line 132
    wsize = 0  # type: longint  # line 133
    to = openIt(saveTo, "w", compress) if saveTo else None  # line 134
    with open(path, "rb") as fd:  # line 135
        while True:  # line 136
            buffer = fd.read(bufSize)  # line 137
            _hash.update(buffer)  # line 138
            if to:  # line 139
                to.write(buffer)  # line 139
            if len(buffer) < bufSize:  # line 140
                break  # line 140
        if to:  # line 141
            to.close()  # line 142
            wsize = os.stat(saveTo).st_size  # line 143
    return (_hash.hexdigest(), wsize)  # line 144

def firstOfMap(map: 'Dict[str, Any]', params: '_coconut.typing.Sequence[str]', default: 'Any'=None) -> 'Any':  # line 146
    ''' Utility. '''  # line 147
    for k, v in map.items():  # line 148
        if k in params:  # line 149
            return v  # line 149
    return default  # line 150

@_coconut_tco  # line 152
def strftime(timestamp: '_coconut.typing.Optional[int]'=None) -> 'str':  # line 152
    return _coconut_tail_call(time.strftime, "%Y-%m-%d %H:%M:%S", time.localtime(timestamp / 1000. if timestamp is not None else None))  # line 152

def detectAndLoad(filename: '_coconut.typing.Optional[str]'=None, content: '_coconut.typing.Optional[bytes]'=None) -> 'Tuple[str, bytes, _coconut.typing.Sequence[str]]':  # line 154
    encoding = None  # type: str  # line 154
    eol = None  # type: bytes  # line 154
    lines = []  # type: _coconut.typing.Sequence[str]  # line 155
    if filename is not None:  # line 156
        with open(filename, "rb") as fd:  # line 157
            content = fd.read()  # line 157
    encoding = (lambda _coconut_none_coalesce_item: sys.getdefaultencoding() if _coconut_none_coalesce_item is None else _coconut_none_coalesce_item)(detectEncoding(content))  # line 158
    eol = eoldet(content)  # line 159
    if filename is not None:  # line 160
        with codecs.open(filename, encoding=encoding) as fd:  # line 161
            lines = safeSplit(fd.read(), (b"\n" if eol is None else eol).decode(encoding))  # line 161
    elif content is not None:  # line 162
        lines = safeSplit(content.decode(encoding), (b"\n" if eol is None else eol).decode(encoding))  # line 163
    else:  # line 164
        return (sys.getdefaultencoding(), b"\n", [])  # line 164
    return (encoding, eol, lines)  # line 165

def diffPathSets(last: 'Dict[str, PathInfo]', diff: 'Dict[str, PathInfo]') -> 'ChangeSet':  # line 167
    ''' Computes a changeset between in-memory and on-disk file lists.
      Additions contains PathInfo for entries in diff
      Deletions contains PathInfo for entries not anymore in diff
      Modifications contains PathInfo for entries changed from last to diff (new state)
  '''  # line 172
    changes = ChangeSet({}, {}, {})  # type: ChangeSet  # line 173
    for path, pinfo in last.items():  # line 174
        if path not in diff:  # now change  changes.deletions[path] = pinfo; continue  # line 175
            continue  # now change  changes.deletions[path] = pinfo; continue  # line 175
        vs = diff[path]  # reference to potentially changed path set  # line 176
        if vs.size is None:  # marked for deletion  # line 177
            changes.deletions[path] = pinfo  # marked for deletion  # line 177
            continue  # marked for deletion  # line 177
        if pinfo.size == None:  # re-added  # line 178
            changes.additions[path] = pinfo  # re-added  # line 178
            continue  # re-added  # line 178
        if pinfo.size != vs.size or pinfo.mtime != vs.mtime or pinfo.hash != vs.hash:  # not need to make hash comparison optional here  # line 179
            changes.modifications[path] = vs  # not need to make hash comparison optional here  # line 179
    for path, pinfo in diff.items():  # added loop  # line 180
        if path not in last:  # line 181
            changes.additions[path] = pinfo  # line 181
    assert not any([path in changes.deletions for path in changes.additions])  # invariant checks  # line 182
    assert not any([path in changes.additions for path in changes.deletions])  # line 183
    return changes  # line 184

try:  # line 186
    DataType = TypeVar("DataType", ChangeSet, MergeBlock, BranchInfo)  # line 186
except:  # Python 2  # line 187
    pass  # Python 2  # line 187

@_coconut_tco  # line 189
def dataCopy(_tipe: 'Type[DataType]', _old: 'DataType', *_args, **_kwargs) -> 'DataType':  # line 189
    r = _old._asdict()  # line 189
    r.update(**_kwargs)  # line 189
    return _coconut_tail_call(makedata, _tipe, *(list(_args) + [r[field] for field in _old._fields]))  # line 189
# TODO test namedtuple instead of instantiated tuple types in "datatype"?

@_coconut_tco  # line 192
def merge(file: '_coconut.typing.Optional[bytes]'=None, into: '_coconut.typing.Optional[bytes]'=None, filename: '_coconut.typing.Optional[str]'=None, intoname: '_coconut.typing.Optional[str]'=None, mergeOperation=MergeOperation.BOTH, conflictResolution=ConflictResolution.ASK, diffOnly: 'bool'=False) -> 'Union[bytes, List[MergeBlock]]':  # line 192
    ''' Merges binary text contents 'file' into file 'into', returning merged result. '''  # line 193
    encoding = None  # type: str  # line 194
    othr = None  # type: _coconut.typing.Sequence[str]  # line 194
    othreol = None  # type: _coconut.typing.Optional[bytes]  # line 194
    curr = None  # type: _coconut.typing.Sequence[str]  # line 194
    curreol = None  # type: _coconut.typing.Optional[bytes]  # line 194
    differ = difflib.Differ()  # line 195
    try:  # load files line-wise and normalize line endings (keep the one of the current file) TODO document  # line 196
        encoding, othreol, othr = detectAndLoad(filename=filename, content=file)  # line 197
        encoding, curreol, curr = detectAndLoad(filename=intoname, content=into)  # line 198
    except Exception as E:  # line 199
        Exit("Cannot merge '%s' into '%s': %r" % (filename, intoname, E))  # line 199
    if None not in [othreol, curreol] and othreol != curreol:  # line 200
        warn("Differing EOL-styles detected during merge. Using current file's style for merge output")  # line 200
    output = list(differ.compare(othr, curr))  # type: List[str]  # from generator expression  # line 201
    debug("Diff output: " + "".join([o.replace("\n", "\\n") for o in output]))  # line 202
    blocks = []  # type: List[MergeBlock]  # merged result in blocks  # line 203
    tmp = []  # type: List[str]  # block lines  # line 204
    last = " "  # line 205
    for no, line in enumerate(output + ["X"]):  # EOF marker (difflib's output will never be "X" alone)  # line 206
        if line[0] == last:  # continue filling consecutive block, no matter what type of block  # line 207
            tmp.append(line[2:])  # continue filling consecutive block, no matter what type of block  # line 207
            continue  # continue filling consecutive block, no matter what type of block  # line 207
        if line == "X":  # EOF marker - perform action for remaining block  # line 208
            if len(tmp) == 0:  # nothing left to do  # line 209
                break  # nothing left to do  # line 209
        if last == " ":  # block is same in both files  # line 210
            blocks.append(MergeBlock(MergeBlockType.KEEP, [line for line in tmp], line=no - len(tmp)))  # line 211
        elif last == "-":  # may be a deletion or replacement, store for later  # line 212
            blocks.append(MergeBlock(MergeBlockType.REMOVE, [line for line in tmp], line=no - len(tmp)))  # line 213
        elif last == "+":  # may be insertion or replacement  # line 214
            blocks.append(MergeBlock(MergeBlockType.INSERT, [line for line in tmp], line=no - len(tmp)))  # line 215
            if len(blocks) >= 2 and len(blocks[-1].lines) == len(blocks[-2].lines):  # replaces previously removed section entirely if any  # line 216
                if len(blocks[-1].lines) >= 2 or (blocks[-1].changes is None and blocks[-2].changes is None):  # full block or no intra-line comment  # line 217
                    blocks[-2] = MergeBlock(MergeBlockType.REPLACE, blocks[-1].lines, line=no - len(tmp) - 1, replaces=blocks[-2])  # remember replaced stuff TODO why -1 necessary?  # line 218
                else:  # may have intra-line modifications  # line 219
                    blocks[-2] = MergeBlock(MergeBlockType.MODIFY, blocks[-1].lines, line=no - len(tmp), replaces=blocks[-2], changes=blocks[-1].changes)  # line 220
                blocks.pop()  # remove TOS  # line 221
        elif last == "?":  # intra-line change comment  # line 222
            ilm = getIntraLineMarkers(tmp[0])  # tipe is one of MergeBlockType 1, 2, 4. # "? " line includes a trailing \n for some reason  # line 223
            blocks[-1] = dataCopy(MergeBlock, blocks[-1], tipe=MergeBlockType.MODIFY, changes=ilm)  # update to MODIFY, otherwise may be REPLACE instead  # line 224
        last = line[0]  # line 225
        tmp[:] = [line[2:]]  # line 226
    debug("Diff blocks: " + repr(blocks))  # line 227
    if diffOnly:  # line 228
        return blocks  # line 228
    output = []  # line 229
    for block in blocks:  # line 230
        if block.tipe == MergeBlockType.KEEP:  # line 231
            output.extend(block.lines)  # line 232
        elif block.tipe == MergeBlockType.INSERT and not (mergeOperation & MergeOperation.REMOVE):  # line 233
            output.extend(block.lines)  # line 234
        elif block.tipe == MergeBlockType.REPLACE:  # complete block replacement  # line 235
            if mergeOperation & MergeOperation.INSERT:  # replaced stuff TODO allow insertion BEFORE and alternatively AFTER?  # line 236
                output.extend(block.replaces.lines)  # replaced stuff TODO allow insertion BEFORE and alternatively AFTER?  # line 236
            if not (mergeOperation & MergeBlockType.REMOVE):  # if remove, don't add replaced lines, otherwise do  # line 237
                output.extend(block.lines)  # if remove, don't add replaced lines, otherwise do  # line 237
        elif block.tipe == MergeBlockType.REMOVE and mergeOperation & MergeOperation.INSERT:  # line 238
            output.extend(block.lines)  # line 239
        elif block.tipe == MergeBlockType.MODIFY:  # line 240
            if block.changes is not None and block.changes.tipe == MergeBlockType.INSERT:  # cannot be REMOVE, because it's always "-" then "+"  # line 241
                output.extend(block.lines if mergeOperation & MergeOperation.INSERT else block.replaces.lines)  # TODO logic?  # line 242
            elif block.replaces.changes is not None and block.replaces.changes.tipe == MergeBlockType.REMOVE:  # line 243
                output.extend(block.lines if mergeOperation & MergeOperation.REMOVE else block.replaces.lines)  # line 244
            elif block.changes is not None and block.changes.tipe == MergeBlockType.MODIFY:  # always both sides modified, but may differ in markers  # line 245
                if conflictResolution == ConflictResolution.THEIRS:  # line 246
                    output.extend(block.replaces.lines)  # line 247
                elif conflictResolution == ConflictResolution.MINE:  # line 248
                    output.extend(block.lines)  # line 249
                elif conflictResolution == ConflictResolution.ASK:  # line 250
                    print(ajoin("THR ", block.replaces.lines, "\n"))  # line 251
                    print(ajoin("MIN ", block.lines, "\n"))  # line 252
                    reso = {"i": ConflictResolution.MINE, "t": ConflictResolution.THEIRS, "m": ConflictResolution.NEXT, "u": None}.get(user_input(" Resolve: *M[I]ne, [T]heirs, [M]erge, [U]ser defined: ").strip().lower(), ConflictResolution.MINE)  # type: _coconut.typing.Optional[int]  # line 253
                    debug("User selected %d" % reso)  # line 254
                    _coconut_match_check = False  # line 255
                    _coconut_match_to = reso  # line 255
                    if _coconut_match_to is None:  # line 255
                        _coconut_match_check = True  # line 255
                    if _coconut_match_check:  # line 255
                        warn("Manual input not implemented yet")  # TODO allow multi-line input (CTRL+D?)  # line 256
                    _coconut_match_check = False  # line 257
                    _coconut_match_to = reso  # line 257
                    if _coconut_match_to == ConflictResolution.MINE:  # line 257
                        _coconut_match_check = True  # line 257
                    if _coconut_match_check:  # line 257
                        debug("Using mine")  # line 258
                        output.extend(block.lines)  # line 259
                    _coconut_match_check = False  # line 260
                    _coconut_match_to = reso  # line 260
                    if _coconut_match_to == ConflictResolution.THEIRS:  # line 260
                        _coconut_match_check = True  # line 260
                    if _coconut_match_check:  # line 260
                        debug("Using theirs")  # line 261
                        output.extend(block.replaces.lines)  # line 262
                    _coconut_match_check = False  # line 263
                    _coconut_match_to = reso  # line 263
                    if _coconut_match_to == ConflictResolution.NEXT:  # line 263
                        _coconut_match_check = True  # line 263
                    if _coconut_match_check:  # line 263
                        warn("Intra-line merge not implemented yet, skipping line(s)")  # line 264
                        output.extend(block.replaces.lines)  # line 265
            else:  # e.g. contains a deletion, but user was asking for insert only??  # line 266
                warn("Investigate this case")  # line 267
                output.extend(block.lines)  # default or not .replaces?  # line 268
    debug("Merge output: " + "; ".join(output))  # line 269
    nl = b"\n" if othreol is None else othreol if curreol is None else curreol  # type: bytes  # line 270
    return _coconut_tail_call(nl.join, [line.encode(encoding) for line in output])  # returning bytes  # line 271
# TODO handle check for more/less lines in found -/+ blocks to find common section and splitting prefix/suffix out

@_coconut_tco  # line 274
def getIntraLineMarkers(line: 'str') -> 'Range':  # line 274
    ''' Return (type, [affected indices]) of "? "-line diff markers ("? " suffix must be removed). '''  # line 275
    if "^" in line:  # line 276
        return _coconut_tail_call(Range, MergeBlockType.MODIFY, [i for i, c in enumerate(line) if c == "^"])  # line 276
    if "+" in line:  # line 277
        return _coconut_tail_call(Range, MergeBlockType.INSERT, [i for i, c in enumerate(line) if c == "+"])  # line 277
    if "-" in line:  # line 278
        return _coconut_tail_call(Range, MergeBlockType.REMOVE, [i for i, c in enumerate(line) if c == "-"])  # line 278
    return _coconut_tail_call(Range, MergeBlockType.KEEP, [])  # line 279

def findSosVcsBase() -> 'Tuple[_coconut.typing.Optional[str], _coconut.typing.Optional[str], _coconut.typing.Optional[str]]':  # line 281
    ''' Attempts to find sos and legacy VCS base folders. '''  # line 282
    debug("Detecting root folders...")  # line 283
    path = os.getcwd()  # type: str  # start in current folder, check parent until found or stopped  # line 284
    vcs = (None, None)  # type: Tuple[_coconut.typing.Optional[str], _coconut.typing.Optional[str]]  # line 285
    while not os.path.exists(os.path.join(path, metaFolder)):  # line 286
        contents = set(os.listdir(path))  # line 287
        vcss = [executable for folder, executable in vcsFolders.items() if folder in contents]  # type: _coconut.typing.Sequence[str]  # determine VCS type from dot folder  # line 288
        choice = None  # type: _coconut.typing.Optional[str]  # line 289
        if len(vcss) > 1:  # line 290
            choice = SVN if SVN in vcss else vcss[0]  # line 291
            warn("Detected more than one parallel VCS checkouts %r. Falling back to '%s'" % (vcss, choice))  # line 292
        elif len(vcss) > 0:  # line 293
            choice = vcss[0]  # line 293
        if not vcs[0] and choice:  # memorize current repo root  # line 294
            vcs = (path, choice)  # memorize current repo root  # line 294
        new = os.path.dirname(path)  # get parent path  # line 295
        if new == path:  # avoid infinite loop  # line 296
            break  # avoid infinite loop  # line 296
        path = new  # line 297
    if os.path.exists(os.path.join(path, metaFolder)):  # found something  # line 298
        if vcs[0]:  # already detected vcs base and command  # line 299
            return (path, vcs[0], vcs[1])  # already detected vcs base and command  # line 299
        sos = path  # line 300
        while True:  # continue search for VCS base  # line 301
            new = os.path.dirname(path)  # get parent path  # line 302
            if new == path:  # no VCS folder found  # line 303
                return (sos, None, None)  # no VCS folder found  # line 303
            path = new  # line 304
            contents = set(os.listdir(path))  # line 305
            vcss = [executable for folder, executable in vcsFolders.items() if folder in contents]  # determine VCS type  # line 306
            choice = None  # line 307
            if len(vcss) > 1:  # line 308
                choice = SVN if SVN in vcss else vcss[0]  # line 309
                warn("Detected more than one parallel VCS checkouts %r. Falling back to '%s'" % (vcss, choice))  # line 310
            elif len(vcss) > 0:  # line 311
                choice = vcss[0]  # line 311
            if choice:  # line 312
                return (sos, path, choice)  # line 312
    return (None, vcs[0], vcs[1])  # line 313

def relativize(root: 'str', path: 'str') -> 'Tuple[str, str]':  # line 315
    ''' Gets relative path for specified file. '''  # line 316
    relpath = os.path.relpath(os.path.dirname(os.path.abspath(path)), root).replace(os.sep, SLASH)  # line 317
    return relpath, os.path.join(relpath, os.path.basename(path)).replace(os.sep, SLASH)  # line 318

def parseOnlyOptions(root: 'str', options: '_coconut.typing.Sequence[str]') -> 'Tuple[_coconut.typing.Optional[Frozenset[str]], _coconut.typing.Optional[Frozenset[str]]]':  # line 320
    cwd = os.getcwd()  # type: str  # line 321
    onlys = []  # type: _coconut.typing.Sequence[str]  # line 322
    excps = []  # type: _coconut.typing.Sequence[str]  # line 322
    index = 0  # type: int  # line 322
    while True:  # line 323
        try:  # line 324
            index = 1 + options.index("--only", index)  # line 325
            onlys.append(options[index])  # line 326
        except:  # line 327
            break  # line 327
    index = 0  # line 328
    while True:  # line 329
        try:  # line 330
            index = 1 + options.index("--except", index)  # line 331
            excps.append(options[index])  # line 332
        except:  # line 333
            break  # line 333
    return (frozenset((relativize(root, o) for o in onlys)) if onlys else None, frozenset((relativize(root, e) for e in excps)) if excps else None)  # line 334

def conditionalIntersection(a: '_coconut.typing.Optional[FrozenSet[str]]', b: 'FrozenSet[str]') -> 'FrozenSet[str]':  # line 336
    return a & b if a else b  # line 336



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# type: ignore

# Compiled with Coconut version 1.3.1 [Dead Parrot]

"""Built-in Coconut utilities."""

# Coconut Header: -------------------------------------------------------------

import sys as _coconut_sys
py_chr, py_filter, py_hex, py_input, py_int, py_map, py_object, py_oct, py_open, py_print, py_range, py_str, py_zip, py_filter, py_reversed, py_enumerate = chr, filter, hex, input, int, map, object, oct, open, print, range, str, zip, filter, reversed, enumerate
class _coconut:
    import collections, copy, functools, imp, itertools, operator, types, weakref
    import pickle
    OrderedDict = collections.OrderedDict
    if _coconut_sys.version_info < (3, 3):
        abc = collections
    else:
        import collections.abc as abc
    Exception, IndexError, KeyError, NameError, TypeError, ValueError, StopIteration, classmethod, dict, enumerate, filter, frozenset, getattr, hasattr, hash, id, int, isinstance, issubclass, iter, len, list, map, min, max, next, object, property, range, reversed, set, slice, str, sum, super, tuple, zip, repr = Exception, IndexError, KeyError, NameError, TypeError, ValueError, StopIteration, classmethod, dict, enumerate, filter, frozenset, getattr, hasattr, hash, id, int, isinstance, issubclass, iter, len, list, map, min, max, next, object, property, range, reversed, set, slice, str, sum, super, tuple, zip, repr
def _coconut_NamedTuple(name, fields):
    return _coconut.collections.namedtuple(name, [x for x, t in fields])
class MatchError(Exception):
    """Pattern-matching error. Has attributes .pattern and .value."""
    __slots__ = ("pattern", "value")
class _coconut_tail_call:
    __slots__ = ("func", "args", "kwargs")
    def __init__(self, func, *args, **kwargs):
        self.func, self.args, self.kwargs = func, args, kwargs
_coconut_tco_func_dict = {}
def _coconut_tco(func):
    @_coconut.functools.wraps(func)
    def tail_call_optimized_func(*args, **kwargs):
        call_func = func
        while True:
            wkref = _coconut_tco_func_dict.get(_coconut.id(call_func))
            if wkref is not None and wkref() is call_func:
                call_func = call_func._coconut_tco_func
            result = call_func(*args, **kwargs)  # pass --no-tco to clean up your traceback
            if not isinstance(result, _coconut_tail_call):
                return result
            call_func, args, kwargs = result.func, result.args, result.kwargs
    tail_call_optimized_func._coconut_tco_func = func
    _coconut_tco_func_dict[_coconut.id(tail_call_optimized_func)] = _coconut.weakref.ref(tail_call_optimized_func)
    return tail_call_optimized_func
def _coconut_igetitem(iterable, index):
    if isinstance(iterable, (_coconut_reversed, _coconut_map, _coconut.filter, _coconut.zip, _coconut_enumerate, _coconut_count, _coconut.abc.Sequence)):
        return iterable[index]
    if not _coconut.isinstance(index, _coconut.slice):
        if index < 0:
            return _coconut.collections.deque(iterable, maxlen=-index)[0]
        return _coconut.next(_coconut.itertools.islice(iterable, index, index + 1))
    if index.start is not None and index.start < 0 and (index.stop is None or index.stop < 0) and index.step is None:
        queue = _coconut.collections.deque(iterable, maxlen=-index.start)
        if index.stop is not None:
            queue = _coconut.tuple(queue)[:index.stop - index.start]
        return queue
    if (index.start is not None and index.start < 0) or (index.stop is not None and index.stop < 0) or (index.step is not None and index.step < 0):
        return _coconut.tuple(iterable)[index]
    return _coconut.itertools.islice(iterable, index.start, index.stop, index.step)
class _coconut_base_compose:
    __slots__ = ("func", "funcstars")
    def __init__(self, func, *funcstars):
        self.func = func
        self.funcstars = []
        for f, star in funcstars:
            if isinstance(f, _coconut_base_compose):
                self.funcstars.append((f.func, star))
                self.funcstars += f.funcstars
            else:
                self.funcstars.append((f, star))
    def __call__(self, *args, **kwargs):
        arg = self.func(*args, **kwargs)
        for f, star in self.funcstars:
            arg = f(*arg) if star else f(arg)
        return arg
    def __repr__(self):
        return _coconut.repr(self.func) + " " + " ".join(("..*> " if star else "..> ") + _coconut.repr(f) for f, star in self.funcstars)
    def __reduce__(self):
        return (self.__class__, (self.func,) + _coconut.tuple(self.funcstars))
def _coconut_forward_compose(func, *funcs): return _coconut_base_compose(func, *((f, False) for f in funcs))
def _coconut_back_compose(*funcs): return _coconut_forward_compose(*_coconut.reversed(funcs))
def _coconut_forward_star_compose(func, *funcs): return _coconut_base_compose(func, *((f, True) for f in funcs))
def _coconut_back_star_compose(*funcs): return _coconut_forward_star_compose(*_coconut.reversed(funcs))
def _coconut_pipe(x, f): return f(x)
def _coconut_star_pipe(xs, f): return f(*xs)
def _coconut_back_pipe(f, x): return f(x)
def _coconut_back_star_pipe(f, xs): return f(*xs)
def _coconut_bool_and(a, b): return a and b
def _coconut_bool_or(a, b): return a or b
def _coconut_none_coalesce(a, b): return a if a is not None else b
def _coconut_minus(a, *rest):
    if not rest:
        return -a
    for b in rest:
        a = a - b
    return a
@_coconut.functools.wraps(_coconut.itertools.tee)
def tee(iterable, n=2):
    if n >= 0 and _coconut.isinstance(iterable, (_coconut.tuple, _coconut.frozenset)):
        return (iterable,) * n
    if n > 0 and (_coconut.hasattr(iterable, "__copy__") or _coconut.isinstance(iterable, _coconut.abc.Sequence)):
        return (iterable,) + _coconut.tuple(_coconut.copy.copy(iterable) for _ in _coconut.range(n - 1))
    return _coconut.itertools.tee(iterable, n)
class reiterable:
    """Allows an iterator to be iterated over multiple times."""
    __slots__ = ("iter",)
    def __init__(self, iterable):
        self.iter = iterable
    def __iter__(self):
        self.iter, out = _coconut_tee(self.iter)
        return _coconut.iter(out)
    def __getitem__(self, index):
        return _coconut_igetitem(_coconut.iter(self), index)
    def __reversed__(self):
        return _coconut_reversed(_coconut.iter(self))
    def __len__(self):
        return _coconut.len(self.iter)
    def __repr__(self):
        return "reiterable(" + _coconut.repr(self.iter) + ")"
    def __reduce__(self):
        return (self.__class__, (self.iter,))
    def __copy__(self):
        return self.__class__(_coconut.copy.copy(self.iter))
    def __fmap__(self, func):
        return _coconut_map(func, self)
class scan:
    """Reduce func over iterable, yielding intermediate results."""
    __slots__ = ("func", "iter")
    def __init__(self, func, iterable):
        self.func, self.iter = func, iterable
    def __iter__(self):
        acc = empty_acc = _coconut.object()
        for item in self.iter:
            if acc is empty_acc:
                acc = item
            else:
                acc = self.func(acc, item)
            yield acc
    def __len__(self):
        return _coconut.len(self.iter)
    def __repr__(self):
        return "scan(" + _coconut.repr(self.iter) + ")"
    def __reduce__(self):
        return (self.__class__, (self.func, self.iter))
    def __copy__(self):
        return self.__class__(self.func, _coconut.copy.copy(self.iter))
    def __fmap__(self, func):
        return _coconut_map(func, self)
class reversed:
    __slots__ = ("_iter",)
    if hasattr(_coconut.map, "__doc__"):
        __doc__ = _coconut.reversed.__doc__
    def __new__(cls, iterable):
        if _coconut.isinstance(iterable, _coconut.range):
            return iterable[::-1]
        if not _coconut.hasattr(iterable, "__reversed__") or _coconut.isinstance(iterable, (_coconut.list, _coconut.tuple)):
            return _coconut.object.__new__(cls)
        return _coconut.reversed(iterable)
    def __init__(self, iterable):
        self._iter = iterable
    def __iter__(self):
        return _coconut.iter(_coconut.reversed(self._iter))
    def __getitem__(self, index):
        if _coconut.isinstance(index, _coconut.slice):
            return _coconut_igetitem(self._iter, _coconut.slice(-(index.start + 1) if index.start is not None else None, -(index.stop + 1) if index.stop else None, -(index.step if index.step is not None else 1)))
        return _coconut_igetitem(self._iter, -(index + 1))
    def __reversed__(self):
        return self._iter
    def __len__(self):
        return _coconut.len(self._iter)
    def __repr__(self):
        return "reversed(" + _coconut.repr(self._iter) + ")"
    def __hash__(self):
        return -_coconut.hash(self._iter)
    def __reduce__(self):
        return (self.__class__, (self._iter,))
    def __copy__(self):
        return self.__class__(_coconut.copy.copy(self._iter))
    def __eq__(self, other):
        return isinstance(other, self.__class__) and self._iter == other._iter
    def __contains__(self, elem):
        return elem in self._iter
    def count(self, elem):
        """Count the number of times elem appears in the reversed iterator."""
        return self._iter.count(elem)
    def index(self, elem):
        """Find the index of elem in the reversed iterator."""
        return _coconut.len(self._iter) - self._iter.index(elem) - 1
    def __fmap__(self, func):
        return self.__class__(_coconut_map(func, self._iter))
class map(_coconut.map):
    __slots__ = ("_func", "_iters")
    if hasattr(_coconut.map, "__doc__"):
        __doc__ = _coconut.map.__doc__
    def __new__(cls, function, *iterables):
        new_map = _coconut.map.__new__(cls, function, *iterables)
        new_map._func, new_map._iters = function, iterables
        return new_map
    def __getitem__(self, index):
        if _coconut.isinstance(index, _coconut.slice):
            return self.__class__(self._func, *(_coconut_igetitem(i, index) for i in self._iters))
        return self._func(*(_coconut_igetitem(i, index) for i in self._iters))
    def __reversed__(self):
        return self.__class__(self._func, *(_coconut_reversed(i) for i in self._iters))
    def __len__(self):
        return _coconut.min(_coconut.len(i) for i in self._iters)
    def __repr__(self):
        return "map(" + _coconut.repr(self._func) + ", " + ", ".join((_coconut.repr(i) for i in self._iters)) + ")"
    def __reduce__(self):
        return (self.__class__, (self._func,) + self._iters)
    def __reduce_ex__(self, _):
        return self.__reduce__()
    def __copy__(self):
        return self.__class__(self._func, *_coconut.map(_coconut.copy.copy, self._iters))
    def __fmap__(self, func):
        return self.__class__(_coconut_forward_compose(self._func, func), *self._iters)
class parallel_map(map):
    """Multi-process implementation of map using concurrent.futures.
    Requires arguments to be pickleable."""
    __slots__ = ()
    def __iter__(self):
        from concurrent.futures import ProcessPoolExecutor
        with ProcessPoolExecutor() as executor:
            return _coconut.iter(_coconut.tuple(executor.map(self._func, *self._iters)))
    def __repr__(self):
        return "parallel_" + _coconut_map.__repr__(self)
class concurrent_map(map):
    """Multi-thread implementation of map using concurrent.futures."""
    __slots__ = ()
    def __iter__(self):
        from concurrent.futures import ThreadPoolExecutor
        from multiprocessing import cpu_count  # cpu_count() * 5 is the default Python 3.5 thread count
        with ThreadPoolExecutor(cpu_count() * 5) as executor:
            return _coconut.iter(_coconut.tuple(executor.map(self._func, *self._iters)))
    def __repr__(self):
        return "concurrent_" + _coconut_map.__repr__(self)
class filter(_coconut.filter):
    __slots__ = ("_func", "_iter")
    if hasattr(_coconut.filter, "__doc__"):
        __doc__ = _coconut.filter.__doc__
    def __new__(cls, function, iterable):
        new_filter = _coconut.filter.__new__(cls, function, iterable)
        new_filter._func, new_filter._iter = function, iterable
        return new_filter
    def __reversed__(self):
        return self.__class__(self._func, _coconut_reversed(self._iter))
    def __repr__(self):
        return "filter(" + _coconut.repr(self._func) + ", " + _coconut.repr(self._iter) + ")"
    def __reduce__(self):
        return (self.__class__, (self._func, self._iter))
    def __reduce_ex__(self, _):
        return self.__reduce__()
    def __copy__(self):
        return self.__class__(self._func, _coconut.copy.copy(self._iter))
    def __fmap__(self, func):
        return _coconut_map(func, self)
class zip(_coconut.zip):
    __slots__ = ("_iters",)
    if hasattr(_coconut.zip, "__doc__"):
        __doc__ = _coconut.zip.__doc__
    def __new__(cls, *iterables):
        new_zip = _coconut.zip.__new__(cls, *iterables)
        new_zip._iters = iterables
        return new_zip
    def __getitem__(self, index):
        if _coconut.isinstance(index, _coconut.slice):
            return self.__class__(*(_coconut_igetitem(i, index) for i in self._iters))
        return _coconut.tuple(_coconut_igetitem(i, index) for i in self._iters)
    def __reversed__(self):
        return self.__class__(*(_coconut_reversed(i) for i in self._iters))
    def __len__(self):
        return _coconut.min(_coconut.len(i) for i in self._iters)
    def __repr__(self):
        return "zip(" + ", ".join((_coconut.repr(i) for i in self._iters)) + ")"
    def __reduce__(self):
        return (self.__class__, self._iters)
    def __reduce_ex__(self, _):
        return self.__reduce__()
    def __copy__(self):
        return self.__class__(*_coconut.map(_coconut.copy.copy, self._iters))
    def __fmap__(self, func):
        return _coconut_map(func, self)
class enumerate(_coconut.enumerate):
    __slots__ = ("_iter", "_start")
    if hasattr(_coconut.enumerate, "__doc__"):
        __doc__ = _coconut.enumerate.__doc__
    def __new__(cls, iterable, start=0):
        new_enumerate = _coconut.enumerate.__new__(cls, iterable, start)
        new_enumerate._iter, new_enumerate._start = iterable, start
        return new_enumerate
    def __getitem__(self, index):
        if _coconut.isinstance(index, _coconut.slice):
            return self.__class__(_coconut_igetitem(self._iter, index), self._start + (0 if index.start is None else index.start if index.start >= 0 else len(self._iter) + index.start))
        return (self._start + index, _coconut_igetitem(self._iter, index))
    def __len__(self):
        return _coconut.len(self._iter)
    def __repr__(self):
        return "enumerate(" + _coconut.repr(self._iter) + ", " + _coconut.repr(self._start) + ")"
    def __reduce__(self):
        return (self.__class__, (self._iter, self._start))
    def __reduce_ex__(self, _):
        return self.__reduce__()
    def __copy__(self):
        return self.__class__(_coconut.copy.copy(self._iter), self._start)
    def __fmap__(self, func):
        return _coconut_map(func, self)
class count:
    """count(start, step) returns an infinite iterator starting at start and increasing by step."""
    __slots__ = ("start", "step")
    def __init__(self, start=0, step=1):
        self.start, self.step = start, step
    def __iter__(self):
        while True:
            yield self.start
            self.start += self.step
    def __contains__(self, elem):
        return elem >= self.start and (elem - self.start) % self.step == 0
    def __getitem__(self, index):
        if _coconut.isinstance(index, _coconut.slice) and (index.start is None or index.start >= 0) and (index.stop is None or index.stop >= 0):
            if index.stop is None:
                return self.__class__(self.start + (index.start if index.start is not None else 0), self.step * (index.step if index.step is not None else 1))
            if _coconut.isinstance(self.start, _coconut.int) and _coconut.isinstance(self.step, _coconut.int):
                return _coconut.range(self.start + self.step * (index.start if index.start is not None else 0), self.start + self.step * index.stop, self.step * (index.step if index.step is not None else 1))
            return _coconut_map(self.__getitem__, _coconut.range(index.start if index.start is not None else 0, index.stop, index.step if index.step is not None else 1))
        if index >= 0:
            return self.start + self.step * index
        raise _coconut.IndexError("count indices must be positive")
    def count(self, elem):
        """Count the number of times elem appears in the count."""
        return int(elem in self)
    def index(self, elem):
        """Find the index of elem in the count."""
        if elem not in self:
            raise _coconut.ValueError(_coconut.repr(elem) + " is not in count")
        return (elem - self.start) // self.step
    def __repr__(self):
        return "count(" + _coconut.str(self.start) + ", " + _coconut.str(self.step) + ")"
    def __hash__(self):
        return _coconut.hash((self.start, self.step))
    def __reduce__(self):
        return (self.__class__, (self.start, self.step))
    def __copy__(self):
        return self.__class__(self.start, self.step)
    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.start == other.start and self.step == other.step
    def __fmap__(self, func):
        return _coconut_map(func, self)
class groupsof:
    """groupsof(n, iterable) splits iterable into groups of size n.
    If the length of the iterable is not divisible by n, the last group may be of size < n."""
    __slots__ = ("group_size", "iter")
    def __init__(self, n, iterable):
        self.iter = iterable
        try:
            self.group_size = _coconut.int(n)
        except _coconut.ValueError:
            raise _coconut.TypeError("group size must be an int; not %r" % (n,))
        if self.group_size <= 0:
            raise _coconut.ValueError("group size must be > 0; not %s" % (self.group_size,))
    def __iter__(self):
        loop, iterator = True, _coconut.iter(self.iter)
        while loop:
            group = []
            for _ in _coconut.range(self.group_size):
                try:
                    group.append(_coconut.next(iterator))
                except _coconut.StopIteration:
                    loop = False
                    break
            if group:
                yield _coconut.tuple(group)
    def __len__(self):
        return _coconut.len(self.iter)
    def __repr__(self):
        return "groupsof(%r)" % (_coconut.repr(self.iter),)
    def __reduce__(self):
        return (self.__class__, (self.group_size, self.iter))
    def __copy__(self):
        return self.__class__(self.group_size, _coconut.copy.copy(self.iter))
    def __fmap__(self, func):
        return _coconut_map(func, self)
def recursive_iterator(func):
    """Decorator that optimizes a function for iterator recursion."""
    tee_store, backup_tee_store = {}, []
    @_coconut.functools.wraps(func)
    def recursive_iterator_func(*args, **kwargs):
        key, use_backup = (args, _coconut.frozenset(kwargs)), False
        try:
            hash(key)
        except _coconut.Exception:
            try:
                key = _coconut.pickle.dumps(key, -1)
            except _coconut.Exception:
                use_backup = True
        if use_backup:
            for i, (k, v) in _coconut.enumerate(backup_tee_store):
                if k == key:
                    to_tee, store_pos = v, i
                    break
            else:  # no break
                to_tee, store_pos = func(*args, **kwargs), None
            to_store, to_return = _coconut_tee(to_tee)
            if store_pos is None:
                backup_tee_store.append([key, to_store])
            else:
                backup_tee_store[store_pos][1] = to_store
        else:
            tee_store[key], to_return = _coconut_tee(tee_store.get(key) or func(*args, **kwargs))
        return to_return
    return recursive_iterator_func
def addpattern(base_func):
    """Decorator to add a new case to a pattern-matching function,
    where the new case is checked last."""
    def pattern_adder(func):
        @_coconut_tco
        @_coconut.functools.wraps(func)
        def add_pattern_func(*args, **kwargs):
            try:
                return base_func(*args, **kwargs)
            except _coconut_MatchError:
                return _coconut_tail_call(func, *args, **kwargs)
        return add_pattern_func
    return pattern_adder
def prepattern(base_func):
    """DEPRECATED: Use addpattern instead."""
    def pattern_prepender(func):
        return addpattern(func)(base_func)
    return pattern_prepender
class _coconut_partial:
    __slots__ = ("func", "_argdict", "_arglen", "_stargs", "keywords")
    if hasattr(_coconut.functools.partial, "__doc__"):
        __doc__ = _coconut.functools.partial.__doc__
    def __init__(self, func, argdict, arglen, *args, **kwargs):
        self.func, self._argdict, self._arglen, self._stargs, self.keywords = func, argdict, arglen, args, kwargs
    def __reduce__(self):
        return (self.__class__, (self.func, self._argdict, self._arglen) + self._stargs, self.keywords)
    def __setstate__(self, keywords):
        self.keywords = keywords
    @property
    def args(self):
        return _coconut.tuple(self._argdict.get(i) for i in _coconut.range(self._arglen)) + self._stargs
    def __call__(self, *args, **kwargs):
        callargs = []
        argind = 0
        for i in _coconut.range(self._arglen):
            if i in self._argdict:
                callargs.append(self._argdict[i])
            elif argind >= _coconut.len(args):
                raise _coconut.TypeError("expected at least " + _coconut.str(self._arglen - _coconut.len(self._argdict)) + " argument(s) to " + _coconut.repr(self))
            else:
                callargs.append(args[argind])
                argind += 1
        callargs += self._stargs
        callargs += args[argind:]
        kwargs.update(self.keywords)
        return self.func(*callargs, **kwargs)
    def __repr__(self):
        args = []
        for i in _coconut.range(self._arglen):
            if i in self._argdict:
                args.append(_coconut.repr(self._argdict[i]))
            else:
                args.append("?")
        for arg in self._stargs:
            args.append(_coconut.repr(arg))
        return _coconut.repr(self.func) + "$(" + ", ".join(args) + ")"
def makedata(data_type, *args, **kwargs):
    """Call the original constructor of the given data type or class with the given arguments."""
    if _coconut.hasattr(data_type, "_make") and (_coconut.issubclass(data_type, _coconut.tuple) or _coconut.isinstance(data_type, _coconut.tuple)):
        return data_type._make(args, **kwargs)
    return _coconut.super(data_type, data_type).__new__(data_type, *args, **kwargs)
def datamaker(data_type):
    """DEPRECATED: Use makedata instead."""
    return _coconut.functools.partial(makedata, data_type)
def consume(iterable, keep_last=0):
    """consume(iterable, keep_last) fully exhausts iterable and return the last keep_last elements."""
    return _coconut.collections.deque(iterable, maxlen=keep_last)  # fastest way to exhaust an iterator
class starmap(_coconut.itertools.starmap):
    __slots__ = ("_func", "_iter")
    if hasattr(_coconut.itertools.starmap, "__doc__"):
        __doc__ = _coconut.itertools.starmap.__doc__
    def __new__(cls, function, iterable):
        new_map = _coconut.itertools.starmap.__new__(cls, function, iterable)
        new_map._func, new_map._iter = function, iterable
        return new_map
    def __getitem__(self, index):
        if _coconut.isinstance(index, _coconut.slice):
            return self.__class__(self._func, _coconut_igetitem(self._iter, index))
        return self._func(*_coconut_igetitem(self._iter, index))
    def __reversed__(self):
        return self.__class__(self._func, *_coconut_reversed(self._iter))
    def __len__(self):
        return _coconut.len(self._iter)
    def __repr__(self):
        return "starmap(" + _coconut.repr(self._func) + ", " + _coconut.repr(self._iter) + ")"
    def __reduce__(self):
        return (self.__class__, (self._func, self._iter))
    def __reduce_ex__(self, _):
        return self.__reduce__()
    def __copy__(self):
        return self.__class__(self._func, _coconut.copy.copy(self._iter))
    def __fmap__(self, func):
        return self.__class__(_coconut_forward_compose(self._func, func), self._iter)
def fmap(func, obj):
    """fmap(func, obj) creates a copy of obj with func applied to its contents.
    Override by defining .__fmap__(func)."""
    if _coconut.hasattr(obj, "__fmap__"):
        return obj.__fmap__(func)
    args = _coconut_starmap(func, obj.items()) if _coconut.isinstance(obj, _coconut.abc.Mapping) else _coconut_map(func, obj)
    if _coconut.isinstance(obj, _coconut.tuple) and _coconut.hasattr(obj, "_make"):
        return obj._make(args)
    if _coconut.isinstance(obj, (_coconut.map, _coconut.range, _coconut.abc.Iterator)):
        return args
    if _coconut.isinstance(obj, _coconut.str):
        return "".join(args)
    return obj.__class__(args)
_coconut_MatchError, _coconut_count, _coconut_enumerate, _coconut_reversed, _coconut_map, _coconut_starmap, _coconut_tee, _coconut_zip, reduce, takewhile, dropwhile = MatchError, count, enumerate, reversed, map, starmap, tee, zip, _coconut.functools.reduce, _coconut.itertools.takewhile, _coconut.itertools.dropwhile
